# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import datetime
from dataclasses import dataclass
from typing import Iterable, Sequence

from astrapy.authentication import (
    EmbeddingAPIKeyHeaderProvider,
    EmbeddingHeadersProvider,
    RerankingAPIKeyHeaderProvider,
    RerankingHeadersProvider,
    StaticTokenProvider,
    TokenProvider,
    coerce_possible_embedding_headers_provider,
    coerce_possible_reranking_headers_provider,
    coerce_possible_token_provider,
)
from astrapy.constants import CallerType, Environment, MapEncodingMode
from astrapy.settings.defaults import (
    API_PATH_ENV_MAP,
    API_VERSION_ENV_MAP,
    DEFAULT_ACCEPT_NAIVE_DATETIMES,
    DEFAULT_BINARY_ENCODE_VECTORS,
    DEFAULT_COLLECTION_ADMIN_TIMEOUT_MS,
    DEFAULT_CUSTOM_DATATYPES_IN_READING,
    DEFAULT_DATABASE_ADMIN_TIMEOUT_MS,
    DEFAULT_DATETIME_TZINFO,
    DEFAULT_ENCODE_MAPS_AS_LISTS_IN_TABLES,
    DEFAULT_GENERAL_METHOD_TIMEOUT_MS,
    DEFAULT_KEYSPACE_ADMIN_TIMEOUT_MS,
    DEFAULT_REQUEST_TIMEOUT_MS,
    DEFAULT_TABLE_ADMIN_TIMEOUT_MS,
    DEFAULT_UNROLL_ITERABLES_TO_LISTS,
    DEFAULT_USE_DECIMALS_IN_COLLECTIONS,
    DEV_OPS_URL_ENV_MAP,
    DEV_OPS_VERSION_ENV_MAP,
    FIXED_SECRET_PLACEHOLDER,
)
from astrapy.utils.unset import _UNSET, UnsetType


@dataclass
class TimeoutOptions:
    """
    The group of settings for the API Options concerning the configured timeouts
    for various kinds of API operations.

    All timeout values are integers expressed
    in milliseconds. A timeout of zero signifies that no timeout is imposed at all
    on that kind of operation.

    All methods that issue HTTP requests allow for a per-invocation override
    of the relevant timeouts involved (see the method docstring and signature
    for details).

    This class is used to override default settings when creating objects such
    as DataAPIClient, Database, Table, Collection and so on. Values that are left
    unspecified will keep the values inherited from the parent "spawner" class.
    See the `APIOptions` master object for more information and usage examples.

    Attributes:
        request_timeout_ms: the timeout imposed on a single HTTP request.
            This is applied to all HTTP requests to both the Data API and the DevOps
            API, with the specific exception of the one request issued by the
            Database's create_collection method (which despite being a single request
            may take a considerable amount of time to complete). Defaults to 10 s.
        general_method_timeout_ms: a timeout to use on the overall duration of a
            method invocation, valid for DML methods which are not concerned with
            schema or admin operations. For some methods (such as `find_one` for
            the Table or Collection classes), this coincides with `request_timeout_ms`:
            in that case, the minimum value of the two is used to limit the request
            duration. For other methods, which possibly comprise several HTTP requests
            (for example `insert_many`), this timeout limits the overall duration time
            of the method invocation, while the per-request timeout is still determined
            by `request_timeout_ms`, separately. Defaults to 30 s.
        collection_admin_timeout_ms: a timeout for all collection-related schema and
            admin operations: creating/dropping a collection, as well as listing them.
            With the exception of collection creation, each individual request issued as
            part of collection-schema operations still must obey `request_timeout_ms`.
            Defaults to 60 s.
        table_admin_timeout_ms: a timeout for all table-related schema and
            admin operations: creating/altering/dropping a table or a table index,
            as well as listing tables/indexes. Each individual request issued as part
            of table-schema operations still must obey `request_timeout_ms`.
            Defaults to 30 s.
        database_admin_timeout_ms: a timeout for all database-related admin operations.
            Creating/dropping a database, listing databases, getting database info,
            querying for the available embedding providers, and others, are all subject
            to this timeout. The longest-running operations in this class are
            the creation and the destruction of a database: if called with the
            `wait_until_complete=True` parameter, these can last several minutes.
            This is the timeout that control if and when the method invocation should
            ever error with an `astrapy.exceptions.DataAPITimeoutException`.
            Note that individual API request issued within those operation still must
            obey the per-request `request_timeout_ms` additional constraint.
            Defaults to 10 m.
        keyspace_admin_timeout_ms: a timeout for all keyspace-related admin operations.
            Creating, altering and dropping a keyspace, as well as listing keyspaces,
            are operations controlled by this timeout. Individual API request issued
            within those operation still must obey the per-request
            `request_timeout_ms` additional constraint. Defaults to 30 s.
    """

    request_timeout_ms: int | UnsetType = _UNSET
    general_method_timeout_ms: int | UnsetType = _UNSET
    collection_admin_timeout_ms: int | UnsetType = _UNSET
    table_admin_timeout_ms: int | UnsetType = _UNSET
    database_admin_timeout_ms: int | UnsetType = _UNSET
    keyspace_admin_timeout_ms: int | UnsetType = _UNSET


@dataclass
class FullTimeoutOptions(TimeoutOptions):
    """
    The group of settings for the API Options concerning the configured timeouts
    for various kinds of API operations.

    All timeout values are integers expressed
    in milliseconds. A timeout of zero signifies that no timeout is imposed at all
    on that kind of operation.

    All methods that issue HTTP requests allow for a per-invocation override
    of the relevant timeouts involved (see the method docstring and signature
    for details).

    This is the "full" version of the class, with the guarantee that all of its members
    have defined values. As such, this is what classes such as DataAPIClient, Database,
    Table, Collection and so on have in their `.api_options` attribute -- as opposed
    to the (non-full) `TimeoutOptions` counterpart class: the latter admits "unset"
    attributes and is used to override specific settings.

    Attributes:
        request_timeout_ms: the timeout imposed on a single HTTP request.
            This is applied to all HTTP requests to both the Data API and the DevOps
            API, with the specific exception of the one request issued by the
            Database's create_collection method (which despite being a single request
            may take a considerable amount of time to complete). Defaults to 10 s.
        general_method_timeout_ms: a timeout to use on the overall duration of a
            method invocation, valid for DML methods which are not concerned with
            schema or admin operations. For some methods (such as `find_one` for
            the Table or Collection classes), this coincides with `request_timeout_ms`:
            in that case, the minimum value of the two is used to limit the request
            duration. For other methods, which possibly comprise several HTTP requests
            (for example `insert_many`), this timeout limits the overall duration time
            of the method invocation, while the per-request timeout is still determined
            by `request_timeout_ms`, separately. Defaults to 30 s.
        collection_admin_timeout_ms: a timeout for all collection-related schema and
            admin operations: creating/dropping a collection, as well as listing them.
            With the exception of collection creation, each individual request issued as
            part of collection-schema operations still must obey `request_timeout_ms`.
            Defaults to 60 s.
        table_admin_timeout_ms: a timeout for all table-related schema and
            admin operations: creating/altering/dropping a table or a table index,
            as well as listing tables/indexes. Each individual request issued as part
            of table-schema operations still must obey `request_timeout_ms`.
            Defaults to 30 s.
        database_admin_timeout_ms: a timeout for all database-related admin operations.
            Creating/dropping a database, listing databases, getting database info,
            querying for the available embedding providers, and others, are all subject
            to this timeout. The longest-running operations in this class are
            the creation and the destruction of a database: if called with the
            `wait_until_complete=True` parameter, these can last several minutes.
            This is the timeout that control if and when the method invocation should
            ever error with an `astrapy.exceptions.DataAPITimeoutException`.
            Note that individual API request issued within those operation still must
            obey the per-request `request_timeout_ms` additional constraint.
            Defaults to 10 m.
        keyspace_admin_timeout_ms: a timeout for all keyspace-related admin operations.
            Creating, altering and dropping a keyspace, as well as listing keyspaces,
            are operations controlled by this timeout. Individual API request issued
            within those operation still must obey the per-request
            `request_timeout_ms` additional constraint. Defaults to 30 s.
    """

    request_timeout_ms: int
    general_method_timeout_ms: int
    collection_admin_timeout_ms: int
    table_admin_timeout_ms: int
    database_admin_timeout_ms: int
    keyspace_admin_timeout_ms: int

    def __init__(
        self,
        *,
        request_timeout_ms: int,
        general_method_timeout_ms: int,
        collection_admin_timeout_ms: int,
        table_admin_timeout_ms: int,
        database_admin_timeout_ms: int,
        keyspace_admin_timeout_ms: int,
    ) -> None:
        TimeoutOptions.__init__(
            self,
            request_timeout_ms=request_timeout_ms,
            general_method_timeout_ms=general_method_timeout_ms,
            collection_admin_timeout_ms=collection_admin_timeout_ms,
            table_admin_timeout_ms=table_admin_timeout_ms,
            database_admin_timeout_ms=database_admin_timeout_ms,
            keyspace_admin_timeout_ms=keyspace_admin_timeout_ms,
        )

    def with_override(self, other: TimeoutOptions) -> FullTimeoutOptions:
        """
        Given an "overriding" set of options, possibly not defined in all its
        attributes, apply the override logic and return a new full options object.

        Args:
            other: a not-necessarily-fully-specified options object. All its defined
                settings take precedence.
        """

        return FullTimeoutOptions(
            request_timeout_ms=(
                other.request_timeout_ms
                if not isinstance(other.request_timeout_ms, UnsetType)
                else self.request_timeout_ms
            ),
            general_method_timeout_ms=(
                other.general_method_timeout_ms
                if not isinstance(other.general_method_timeout_ms, UnsetType)
                else self.general_method_timeout_ms
            ),
            collection_admin_timeout_ms=(
                other.collection_admin_timeout_ms
                if not isinstance(other.collection_admin_timeout_ms, UnsetType)
                else self.collection_admin_timeout_ms
            ),
            table_admin_timeout_ms=(
                other.table_admin_timeout_ms
                if not isinstance(other.table_admin_timeout_ms, UnsetType)
                else self.table_admin_timeout_ms
            ),
            database_admin_timeout_ms=(
                other.database_admin_timeout_ms
                if not isinstance(other.database_admin_timeout_ms, UnsetType)
                else self.database_admin_timeout_ms
            ),
            keyspace_admin_timeout_ms=(
                other.keyspace_admin_timeout_ms
                if not isinstance(other.keyspace_admin_timeout_ms, UnsetType)
                else self.keyspace_admin_timeout_ms
            ),
        )


@dataclass
class SerdesOptions:
    """
    The group of settings for the API Options concerning serialization and
    deserialization of values to and from the Data API. Write-path settings
    affect how values are encoded in the JSON payload to API requests, and
    read-path settings determine the choice of data types used to represent
    values found in the API responses when a method returns data.

    This class is used to override default settings when creating objects such
    as DataAPIClient, Database, Table, Collection and so on. Values that are left
    unspecified will keep the values inherited from the parent "spawner" class.
    See the `APIOptions` master object for more information and usage examples.

    Only the classes that directly exchange data with the API, i.e. Table and
    Collection, make actual use of the settings in this group. Nevertheless,
    all objects in the hierarchy (for example DataAPIClient) have their own
    customizable SerdesOptions settings: this makes it easy to configure one's
    preference once and just have each spawned Collection or Table inherit the
    desired settings.

    Attributes:
        binary_encode_vectors: Write-Path. Whether to encode vectors using the faster,
            more efficient binary encoding as opposed to sending plain lists
            of numbers. For Tables, this affects vectors passed to write methods
            as instances of `DataAPIVector`, while for collections this affects
            the encoding of the quantity found in the "$vector" field, if present,
            regardless of its representation in the method argument. Defaults to True.
        custom_datatypes_in_reading: Read-Path. This setting determines whether return
            values from read methods should use astrapy custom classes (default setting
            of True), or try to use only standard-library data types instead (False).
            The astrapy classes are designed to losslessly express DB values.
            For Collections, this setting only affects timestamp and vector values,
            while when reading from a Table there are several other implications.
            Keep in mind that for some of the data types, choosing the stdlib fallback
            may lead to approximate results and even lead to errors (for instance if a
            date stored on database falls outside of the range the stdlib can express).
            Here is a list of the fallbacks (see the individual classes for more info):
            * DataAPIVector => list[float]: no loss of expressivity.
            * DataAPITimestamp  => datetime.datetime: shorter year range (1AD-9999AD)
            and lossy storage of sub-second part for ancient years.
            * DataAPIDate => datetime.date: same year range limitations.
            * DataAPITime => datetime.time: approximation (datetime.time
            has microsecond precision, so the nanosecond part, if present, is lost).
            * DataAPIDuration => datetime.timedelta: durations are intrinsically a
            different thing (see `DataAPIDuration` class). If requested, a coercion
            into a timedelta is attempted, but (a) months are not expressable, (b)
            days are always interpreted as made of 24 hours (despite the occasional
            23- or 25-hour day), and (c) nanoseconds are lost just like for DataAPITime.
            * DataAPIMap => dict: would raise an error if non-hashable data types (e.g.
            lists) are allowed as keys in map-columns on the Table.
            * DataAPISet => set: would raise an error if non-hashable data types (e.g.
            lists) are allowed as entries in set-columns on the Table.
        unroll_iterables_to_lists: Write-Path. If this is set to True, a wider group
            of data types can be provided where a list of values is expected: anything
            that can be iterated over is automatically "unrolled" into a list prior to
            insertion. This means, for example, that one can directly use classes such
            as numpy objects or generators for writes where a vector is expected.
            This setting defaults to False, as it incurs some performance cost (due
            to preemptive inspection of all values in dictionaries provided for writes).
        use_decimals_in_collections: both Read- and Write-path. The `decimal.Decimal`
            standard library class represents lossless numbers with exact arithmetic
            (as opposed to the standard Python floats which hold a finite number
            of significant digits, thus possibly with approximate arithmetic).
            This settings, relevant for Collections only, affects both paths:
            * for the Write-Path, if set to True, `Decimal` instances are accepted
            for storage into collections and are written losslessly. The default
            value of False means that `Decimal` numbers are not accepted for writes.
            * for the Read-Path, if this is set to True, then *all* numeric values
            found in documents from the collection (except the values in "$vector")
            are returned as `Decimal` instances. Conversely, with the default setting
            of `use_decimals_in_collections = False`, read documents are returned as
            containing only regular integers and floats.
            Before switching this setting to True, one should consider the actual need
            for lossless arbitrary decimal precision in their application: besides the
            fact that every number is then returned as an instance of `Decimal` when
            reading from collections, an additional performance cost is required to
            manage the serialization/deserialization of objects exchanged with the API.
        encode_maps_as_lists_in_tables: Write-Path. Whether 'maps' (`dict` and/or
            `DataAPIMap` objects alike) are automatically converted into association
            lists if they are found in suitable portions of the write payload for Table
            operations, i.e. if they represent parts of a row (including updates, etc).
            Takes values in the `astrapy.constants.MapEncodingMode` enum: for "NEVER",
            no such conversion occurs; for "DATAAPIMAPS" it takes place only for
            instances of `DataAPIMap`, for "ALWAYS" also `dict` objects are transformed.
            When (and where) enabled, the automatic conversions have the following form:
            * `{10: "ten"} ==> [[10, "ten"]]`
            * `DataAPIMap([('a', 1), ('b', 2)]) ==> [["a", 1], ["b", 2]]`.
            Defaults to "NEVER".
        accept_naive_datetimes: Write-Path. Python datetimes can be either "naive" or
            "aware" of a timezone/offset information. Only the latter type can be
            translated unambiguously and without implied assumptions into a well-defined
            timestamp. Because the Data API always stores timestamps, by default astrapy
            will raise an error if a write is attempted that uses a naive datetime.
            If this setting is changed to True (from its default of False), then astrapy
            will stop complaining about naive datetimes and accept them as valid
            timestamps for writes. These will be converted into timestamps using
            their `.timestamp()` method, which uses the system locale implicitly.
            It is important to appreciate the possible consequences, for example
            if a table or collection is shared by instances of the application
            running with different system locales.
        datetime_tzinfo: Read-Path. When reading timestamps from tables or collection
            with the setting `custom_datatypes_in_reading = False`, ordinary
            `datetime.datetime` objects are returned for timestamps read from the
            database. This setting (defaulting to `datetime.timezone.utc`) determines
            the timezone used in the returned datetime objects. Setting this value
            to None results in naive datetimes being returned (not recommended).
    """

    binary_encode_vectors: bool | UnsetType
    custom_datatypes_in_reading: bool | UnsetType
    unroll_iterables_to_lists: bool | UnsetType
    use_decimals_in_collections: bool | UnsetType
    encode_maps_as_lists_in_tables: MapEncodingMode | UnsetType
    accept_naive_datetimes: bool | UnsetType
    datetime_tzinfo: datetime.timezone | None | UnsetType

    def __init__(
        self,
        *,
        binary_encode_vectors: bool | UnsetType = _UNSET,
        custom_datatypes_in_reading: bool | UnsetType = _UNSET,
        unroll_iterables_to_lists: bool | UnsetType = _UNSET,
        use_decimals_in_collections: bool | UnsetType = _UNSET,
        encode_maps_as_lists_in_tables: str | MapEncodingMode | UnsetType = _UNSET,
        accept_naive_datetimes: bool | UnsetType = _UNSET,
        datetime_tzinfo: datetime.timezone | None | UnsetType = _UNSET,
    ) -> None:
        self.binary_encode_vectors = binary_encode_vectors
        self.custom_datatypes_in_reading = custom_datatypes_in_reading
        self.unroll_iterables_to_lists = unroll_iterables_to_lists
        self.use_decimals_in_collections = use_decimals_in_collections
        if isinstance(encode_maps_as_lists_in_tables, str):
            self.encode_maps_as_lists_in_tables = MapEncodingMode.coerce(
                encode_maps_as_lists_in_tables
            )
        else:
            self.encode_maps_as_lists_in_tables = encode_maps_as_lists_in_tables
        self.accept_naive_datetimes = accept_naive_datetimes
        self.datetime_tzinfo = datetime_tzinfo


@dataclass
class FullSerdesOptions(SerdesOptions):
    """
    The group of settings for the API Options concerning serialization and
    deserialization of values to and from the Data API. Write-path settings
    affect how values are encoded in the JSON payload to API requests, and
    read-path settings determine the choice of data types used to represent
    values found in the API responses when a method returns data.

    This is the "full" version of the class, with the guarantee that all of its members
    have defined values. As such, this is what classes such as DataAPIClient, Database,
    Table, Collection and so on have in their `.api_options` attribute -- as opposed
    to the (non-full) `SerdesOptions` counterpart class: the latter admits "unset"
    attributes and is used to override specific settings.

    Only the classes that directly exchange data with the API, i.e. Table and
    Collection, make actual use of the settings in this group. Nevertheless,
    all objects in the hierarchy (for example DataAPIClient) have their own
    customizable SerdesOptions settings: this makes it easy to configure one's
    preference once and just have each spawned Collection or Table inherit the
    desired settings.

    Attributes:
        binary_encode_vectors: Write-Path. Whether to encode vectors using the faster,
            more efficient binary encoding as opposed to sending plain lists
            of numbers. For Tables, this affects vectors passed to write methods
            as instances of `DataAPIVector`, while for collections this affect
            the encoding of the quantity found in the "$vector" field, if present,
            regardless of its representation in the method argument. Defaults to True.
        custom_datatypes_in_reading: Read-Path. This setting determines whether return
            values from read methods should use astrapy custom classes (default setting
            of True), or try to use only standard-library data types instead (False).
            The astrapy classes are designed to losslessly express DB values.
            For Collections, this setting only affects timestamp and vector values,
            while when reading from a Table there are several other implications.
            Keep in mind that for some of the data types, choosing the stdlib fallback
            may lead to approximate results and even lead to errors (for instance if a
            date stored on database falls outside of the range the stdlib can express).
            Here is a list of the fallbacks (see the individual classes for more info):
            * DataAPIVector => list[float]: no loss of expressivity.
            * DataAPITimestamp  => datetime.datetime: shorter year range (1AD-9999AD)
            and lossy storage of sub-second part for ancient years.
            * DataAPIDate => datetime.date: same year range limitations.
            * DataAPITime => datetime.time: approximation (datetime.time
            has microsecond precision, so the nanosecond part, if present, is lost).
            * DataAPIDuration => datetime.timedelta: durations are intrinsically a
            different thing (see `DataAPIDuration` class). If requested, a coercion
            into a timedelta is attempted, but (a) months are not expressable, (b)
            days are always interpreted as made of 24 hours (despite the occasional
            23- or 25-hour day), and (c) nanoseconds are lost just like for DataAPITime.
            * DataAPIMap => dict: would raise an error if non-hashable data types (e.g.
            lists) are allowed as keys in map-columns on the Table.
            * DataAPISet => set: would raise an error if non-hashable data types (e.g.
            lists) are allowed as entries in set-columns on the Table.
        unroll_iterables_to_lists: Write-Path. If this is set to True, a wider group
            of data types can be provided where a list of values is expected: anything
            that can be iterated over is automatically "unrolled" into a list prior to
            insertion. This means, for example, that one can directly use classes such
            as numpy objects or generators for writes where a vector is expected.
            This setting defaults to False, as it incurs some performance cost (due
            to preemptive inspection of all values in dictionaries provided for writes).
        use_decimals_in_collections: both Read- and Write-path. The `decimal.Decimal`
            standard library class represents lossless numbers with exact arithmetic
            (as opposed to the standard Python floats which hold a finite number
            of significant digits, thus possibly with approximate arithmetic).
            This settings, relevant for Collections only, affects both paths:
            * for the Write-Path, if set to True, `Decimal` instances are accepted
            for storage into collections and are written losslessly. The default
            value of False means that `Decimal` numbers are not accepted for writes.
            * for the Read-Path, if this is set to True, then *all* numeric values
            found in documents from the collection (except the values in "$vector")
            are returned as `Decimal` instances. Conversely, with the default setting
            of `use_decimals_in_collections = False`, read documents are returned as
            containing only regular integers and floats.
            Before switching this setting to True, one should consider the actual need
            for lossless arbitrary decimal precision in their application: besides the
            fact that every number is then returned as an instance of `Decimal` when
            reading from collections, an additional performance cost is required to
            manage the serialization/deserialization of objects exchanged with the API.
        encode_maps_as_lists_in_tables: Write-Path. Whether 'maps' (`dict` and/or
            `DataAPIMap` objects alike) are automatically converted into association
            lists if they are found in suitable portions of the write payload for Table
            operations, i.e. if they represent parts of a row (including updates, etc).
            Takes values in the `astrapy.constants.MapEncodingMode` enum: for "NEVER",
            no such conversion occurs; for "DATAAPIMAPS" it takes place only for
            instances of `DataAPIMap`, for "ALWAYS" also `dict` objects are transformed.
            When (and where) enabled, the automatic conversions have the following form:
            * `{10: "ten"} ==> [[10, "ten"]]`
            * `DataAPIMap([('a', 1), ('b', 2)]) ==> [["a", 1], ["b", 2]]`.
            Defaults to "NEVER".
        accept_naive_datetimes: Write-Path. Python datetimes can be either "naive" or
            "aware" of a timezone/offset information. Only the latter type can be
            translated unambiguously and without implied assumptions into a well-defined
            timestamp. Because the Data API always stores timestamps, by default astrapy
            will raise an error if a write is attempted that uses a naive datetime.
            If this setting is changed to True (from its default of False), then astrapy
            will stop complaining about naive datetimes and accept them as valid
            timestamps for writes. These will be converted into timestamps using
            their `.timestamp()` method, which uses the system locale implicitly.
            It is important to appreciate the possible consequences, for example
            if a table or collection is shared by instances of the application
            running with different system locales.
        datetime_tzinfo: Read-Path. When reading timestamps from tables or collection
            with the setting `custom_datatypes_in_reading = False`, ordinary
            `datetime.datetime` objects are returned for timestamps read from the
            database. This setting (defaulting to `datetime.timezone.utc`) determines
            the timezone used in the returned datetime objects. Setting this value
            to None results in naive datetimes being returned (not recommended).
    """

    binary_encode_vectors: bool
    custom_datatypes_in_reading: bool
    unroll_iterables_to_lists: bool
    use_decimals_in_collections: bool
    encode_maps_as_lists_in_tables: MapEncodingMode
    accept_naive_datetimes: bool
    datetime_tzinfo: datetime.timezone | None

    def __init__(
        self,
        *,
        binary_encode_vectors: bool,
        custom_datatypes_in_reading: bool,
        unroll_iterables_to_lists: bool,
        use_decimals_in_collections: bool,
        encode_maps_as_lists_in_tables: str | MapEncodingMode,
        accept_naive_datetimes: bool,
        datetime_tzinfo: datetime.timezone | None,
    ) -> None:
        SerdesOptions.__init__(
            self,
            binary_encode_vectors=binary_encode_vectors,
            custom_datatypes_in_reading=custom_datatypes_in_reading,
            unroll_iterables_to_lists=unroll_iterables_to_lists,
            use_decimals_in_collections=use_decimals_in_collections,
            encode_maps_as_lists_in_tables=encode_maps_as_lists_in_tables,
            accept_naive_datetimes=accept_naive_datetimes,
            datetime_tzinfo=datetime_tzinfo,
        )

    def with_override(self, other: SerdesOptions) -> FullSerdesOptions:
        """
        Given an "overriding" set of options, possibly not defined in all its
        attributes, apply the override logic and return a new full options object.

        Args:
            other: a not-necessarily-fully-specified options object. All its defined
                settings take precedence.
        """

        return FullSerdesOptions(
            binary_encode_vectors=(
                other.binary_encode_vectors
                if not isinstance(other.binary_encode_vectors, UnsetType)
                else self.binary_encode_vectors
            ),
            custom_datatypes_in_reading=(
                other.custom_datatypes_in_reading
                if not isinstance(other.custom_datatypes_in_reading, UnsetType)
                else self.custom_datatypes_in_reading
            ),
            unroll_iterables_to_lists=(
                other.unroll_iterables_to_lists
                if not isinstance(other.unroll_iterables_to_lists, UnsetType)
                else self.unroll_iterables_to_lists
            ),
            use_decimals_in_collections=(
                other.use_decimals_in_collections
                if not isinstance(other.use_decimals_in_collections, UnsetType)
                else self.use_decimals_in_collections
            ),
            encode_maps_as_lists_in_tables=(
                other.encode_maps_as_lists_in_tables
                if not isinstance(other.encode_maps_as_lists_in_tables, UnsetType)
                else self.encode_maps_as_lists_in_tables
            ),
            accept_naive_datetimes=(
                other.accept_naive_datetimes
                if not isinstance(other.accept_naive_datetimes, UnsetType)
                else self.accept_naive_datetimes
            ),
            datetime_tzinfo=(
                other.datetime_tzinfo
                if not isinstance(other.datetime_tzinfo, UnsetType)
                else self.datetime_tzinfo
            ),
        )


@dataclass
class DataAPIURLOptions:
    """
    The group of settings for the API Options that determines the URL used to
    reach the Data API.

    This class is used to override default settings when creating objects such
    as DataAPIClient, Database, Table, Collection and so on. Values that are left
    unspecified will keep the values inherited from the parent "spawner" class.
    See the `APIOptions` master object for more information and usage examples.
    Keep in mind, However, that only in very specific customized scenarios should
    it be necessary to override the default settings for this class.

    Attributes:
        api_path: path to append to the API Endpoint. For Astra DB environments,
            this can be left to its default of "/api/json"; for other environment,
            the default of "" is correct (unless there are specific redirects in place).
        api_version: version specifier to append to the API path. The default values
            of "v1" should not be changed except specific redirects are in place on
            non-Astra environments.
    """

    api_path: str | None | UnsetType = _UNSET
    api_version: str | None | UnsetType = _UNSET


@dataclass
class FullDataAPIURLOptions(DataAPIURLOptions):
    """
    The group of settings for the API Options that determines the URL used to
    reach the Data API.

    This is the "full" version of the class, with the guarantee that all of its members
    have defined values. As such, this is what classes such as DataAPIClient, Database,
    Table, Collection and so on have in their `.api_options` attribute -- as opposed
    to the (non-full) `DataAPIURLOptions` counterpart class: the latter admits "unset"
    attributes and is used to override specific settings. However, only in very specific
    customized scenarios should it be necessary to override the default settings.

    Attributes:
        api_path: path to append to the API Endpoint. For Astra DB environments,
            this can be left to its default of "/api/json"; for other environment,
            the default of "" is correct (unless there are specific redirects in place).
        api_version: version specifier to append to the API path. The default values
            of "v1" should not be changed except specific redirects are in place on
            non-Astra environments.
    """

    api_path: str | None
    api_version: str | None

    def __init__(
        self,
        *,
        api_path: str | None,
        api_version: str | None,
    ) -> None:
        DataAPIURLOptions.__init__(
            self,
            api_path=api_path,
            api_version=api_version,
        )

    def with_override(self, other: DataAPIURLOptions) -> FullDataAPIURLOptions:
        """
        Given an "overriding" set of options, possibly not defined in all its
        attributes, apply the override logic and return a new full options object.

        Args:
            other: a not-necessarily-fully-specified options object. All its defined
                settings take precedence.
        """

        return FullDataAPIURLOptions(
            api_path=(
                other.api_path
                if not isinstance(other.api_path, UnsetType)
                else self.api_path
            ),
            api_version=(
                other.api_version
                if not isinstance(other.api_version, UnsetType)
                else self.api_version
            ),
        )


@dataclass
class DevOpsAPIURLOptions:
    """
    The group of settings for the API Options that determines the URL used to
    reach the DevOps API.

    This class is used to override default settings when creating objects such
    as DataAPIClient, Database, Table, Collection and so on. Values that are left
    unspecified will keep the values inherited from the parent "spawner" class.
    See the `APIOptions` master object for more information and usage examples.

    The default settings for this class depend on the environment; in particular,
    for non-Astra environments (such as HCD or DSE) there is no DevOps API, hence
    these settings are irrelevant.

    Attributes:
        dev_ops_url: This can be used to specify the URL to the DevOps API. The default
            for production Astra DB is "https://api.astra.datastax.com" and should
            never need to be overridden.
        dev_ops_api_version: this specifies a version for the DevOps API
            (the default is "v2"). It should never need to be overridden.
    """

    dev_ops_url: str | UnsetType = _UNSET
    dev_ops_api_version: str | None | UnsetType = _UNSET


@dataclass
class FullDevOpsAPIURLOptions(DevOpsAPIURLOptions):
    """
    The group of settings for the API Options that determines the URL used to
    reach the DevOps API.

    This is the "full" version of the class, with the guarantee that all of its members
    have defined values. As such, this is what classes such as DataAPIClient, Database,
    Table, Collection and so on have in their `.api_options` attribute -- as opposed
    to the (non-full) `DevOpsAPIURLOptions` counterpart class: the latter admits "unset"
    attributes and is used to override specific settings. However, only in very specific
    customized scenarios should it be necessary to override the default settings.

    The default settings for this class depend on the environment; in particular,
    for non-Astra environments (such as HCD or DSE) there is no DevOps API, hence
    these settings are irrelevant.

    Attributes:
        dev_ops_url: This can be used to specify the URL to the DevOps API. The default
            for production Astra DB is "https://api.astra.datastax.com" and should
            never need to be overridden.
        dev_ops_api_version: this specifies a version for the DevOps API
            (the default is "v2"). It should never need to be overridden.
    """

    dev_ops_url: str
    dev_ops_api_version: str | None

    def __init__(
        self,
        *,
        dev_ops_url: str,
        dev_ops_api_version: str | None,
    ) -> None:
        DevOpsAPIURLOptions.__init__(
            self,
            dev_ops_url=dev_ops_url,
            dev_ops_api_version=dev_ops_api_version,
        )

    def with_override(self, other: DevOpsAPIURLOptions) -> FullDevOpsAPIURLOptions:
        """
        Given an "overriding" set of options, possibly not defined in all its
        attributes, apply the override logic and return a new full options object.

        Args:
            other: a not-necessarily-fully-specified options object. All its defined
                settings take precedence.
        """

        return FullDevOpsAPIURLOptions(
            dev_ops_url=(
                other.dev_ops_url
                if not isinstance(other.dev_ops_url, UnsetType)
                else self.dev_ops_url
            ),
            dev_ops_api_version=(
                other.dev_ops_api_version
                if not isinstance(other.dev_ops_api_version, UnsetType)
                else self.dev_ops_api_version
            ),
        )


@dataclass
class APIOptions:
    """
    This class represents all settings that can be configured for how astrapy
    interacts with the API (both Data API and DevOps API). Each object in the
    abstraction hierarchy (DataAPIClient, Database, Table, Collection, ...)
    has a full set of these options that determine how it behaves when performing
    actions toward the API.

    In order to customize the behavior from its preset defaults, one should create
    an `APIOptions` object and either:
    (a) pass it as the `api_options` argument to the DataAPIClient constructor or
    any of the `.with_options` and `.to_[a]sync` methods, to get a new instance of
    an object with some settings changed.
    (b) pass it as the `spawn_api_options` argument to "spawning methods", such as
    `get/create_collection`, `get/create_table`, `get/create_database` or
    `get_database_admin`, to set these option overrides to the returned object.
    The APIOptions object passed as argument can define zero, some or all of its
    members, overriding the corresponding settings and keeping, for all unspecified
    settings, the values inherited from the object whose method is invoked (see some
    examples below).
    Some of these methods admit a few shorthands for settings to be passed directly,
    such as `token="..."`, as an alternative to wrapping the setting into its own
    APIOptions object: this however, while taking precedence over the full object
    if provided at the same time, only covers the few most important settings one
    may want to customize when spawning objects.

    With the exception of the "database/admin additional headers" attributes and
    the "redacted header names", which are merged with the inherited ones,
    the override logic is the following: if an override is provided (even if it
    is None), it completely replaces the inherited value.

    The structure of APIOptions is one and the same throughout the object hierarchy:
    that makes it possible to set, for example, a serialization option for reading
    from Collections at the Database level, so that each Collection spawned from it
    will have the desired behavior; similarly, this makes it possible to set a
    DevOps-API-related option (such as `dev_ops_api_url_options.dev_ops_url`)
    for a Table, an action which has no effect whatsoever.

    Attributes:
        environment: an identifier for the environment for the Data API. This can
            describe an Astra DB environment (such as the default of "prod"), or
            a self-deployed setup (such as "dse" or "hcd"). This setting cannot be
            overridden through customization: it can only be provided when creating
            the DataAPIClient top object in the abstraction hierarchy.
        callers: an iterable of "caller identities" to be used in identifying the
            caller, through the User-Agent header, when issuing requests to the
            Data API. Each caller identity is a `(name, version)` 2-item tuple whose
            elements can be strings or None.
        database_additional_headers: free-form dictionary of additional headers to
            employ when issuing requests to the Data API from Database, Table and
            Collection classes. Passing a key with a value of None means that a certain
            header is suppressed when issuing the request.
        admin_additional_headers: free-form dictionary of additional headers to
            employ when issuing requests to both the Data API and the DevOps API
            from AstraDBAdmin, AstraDBDatabaseAdmin and DataAPIDatabaseAdmin classes.
            Passing a key with a value of None means that a certain header is
            suppressed when issuing the request.
        redacted_header_names: A set of (case-insensitive) strings denoting the headers
            that contain secrets, thus are to be masked when logging request details.
        token: an instance of TokenProvider to provide authentication to requests.
            Passing a string, or None, to this constructor parameter will get it
            automatically converted into the appropriate TokenProvider object.
            Depending on the target (Data API or DevOps API), this one attribute
            is encoded in the request appropriately.
        embedding_api_key: an instance of EmbeddingHeadersProvider should it be needed
            for vectorize-related data operations (used by Tables and Collections).
            Passing a string, or None, to this constructor parameter will get it
            automatically converted into the appropriate EmbeddingHeadersProvider
            object.
        reranking_api_key: an instance of RerankingHeadersProvider should it be needed
            for reranking-related data operations (used by Tables and Collections).
            Passing a string, or None, to this constructor parameter will get it
            automatically converted into the appropriate RerankingHeadersProvider
            object.
        timeout_options: an instance of `TimeoutOptions` (see) to control the timeout
            behavior for the various kinds of operations involving the Data/DevOps API.
        serdes_options: an instance of `SerdesOptions` (see) to customize the
            serializing/deserializing behavior related to writing to and reading from
            tables and collections.
        data_api_url_options: an instance of `DataAPIURLOptions` (see) to customize
            the full URL used to reach the Data API (customizing this setting
            is rarely needed).
        dev_ops_api_url_options: an instance of `DevOpsAPIURLOptions` (see) to
            customize the URL used to reach the DevOps API (customizing this setting
            is rarely needed; relevant only for Astra DB environments).

    Examples:
            >>> from astrapy import DataAPIClient
            >>> from astrapy.api_options import (
            ...     APIOptions,
            ...     SerdesOptions,
            ...     TimeoutOptions,
            ... )
            >>> from astrapy.authentication import (
            ...     StaticTokenProvider,
            ...     AWSEmbeddingHeadersProvider,
            ... )
            >>>
            >>> # Disable custom datatypes in all reads:
            >>> no_cdt_options = APIOptions(
            ...     serdes_options=SerdesOptions(
            ...         custom_datatypes_in_reading=False,
            ...     )
            ... )
            >>> my_client = DataAPIClient(api_options=no_cdt_options)
            >>>
            >>> # These spawned objects inherit that setting:
            >>> my_database = my_client.get_database(
            ...     "https://...",
            ...     token="my-token-1",
            ... )
            >>> my_table = my_database.get_table("my_table")
            >>>
            >>> # Make a copy of `table` with some redefined timeouts
            >>> # and a certain header-based authentication for its vectorize provider:
            >>> my_table_timeouts = TimeoutOptions(
            ...     request_timeout_ms=15000,
            ...     general_method_timeout_ms=30000,
            ...     table_admin_timeout_ms=120000,
            ... )
            >>> my_table_apikey_provider = AWSEmbeddingHeadersProvider(
            ...     embedding_access_id="my-access-id",
            ...     embedding_secret_id="my-secret-id",
            ... )
            >>> my_table_slow_copy = my_table.with_options(
            ...     api_options=APIOptions(
            ...         embedding_api_key=my_table_apikey_provider,
            ...         timeout_options=my_table_timeouts,
            ...     ),
            ... )
            >>>
            >>> # Create another 'Database' with a different auth token
            >>> # (for get_database, the 'token=' shorthand shown above does the same):
            >>> my_other_database = my_client.get_database(
            ...     "https://...",
            ...     spawn_api_options=APIOptions(
            ...         token="my-token-2",
            ...     ),
            ... )
            >>>
            >>> # Spawn a collection from a database and set it to use
            >>> # another token and a different policy with Decimals:
            >>> my_other_table = my_database.get_collection(
            ...     "my_other_table",
            ...     spawn_api_options=APIOptions(
            ...         token="my-token-3",
            ...         serdes_options=SerdesOptions(
            ...             use_decimals_in_collections=True,
            ...         )
            ...     ),
            ... )
    """

    environment: str | UnsetType = _UNSET
    callers: Sequence[CallerType] | UnsetType = _UNSET
    database_additional_headers: dict[str, str | None] | UnsetType = _UNSET
    admin_additional_headers: dict[str, str | None] | UnsetType = _UNSET
    redacted_header_names: set[str] | UnsetType = _UNSET
    token: TokenProvider | UnsetType = _UNSET
    embedding_api_key: EmbeddingHeadersProvider | UnsetType = _UNSET
    reranking_api_key: RerankingHeadersProvider | UnsetType = _UNSET

    timeout_options: TimeoutOptions | UnsetType = _UNSET
    serdes_options: SerdesOptions | UnsetType = _UNSET
    data_api_url_options: DataAPIURLOptions | UnsetType = _UNSET
    dev_ops_api_url_options: DevOpsAPIURLOptions | UnsetType = _UNSET

    def __init__(
        self,
        *,
        callers: Sequence[CallerType] | UnsetType = _UNSET,
        database_additional_headers: dict[str, str | None] | UnsetType = _UNSET,
        admin_additional_headers: dict[str, str | None] | UnsetType = _UNSET,
        redacted_header_names: Iterable[str] | UnsetType = _UNSET,
        token: str | TokenProvider | UnsetType = _UNSET,
        embedding_api_key: str | EmbeddingHeadersProvider | UnsetType = _UNSET,
        reranking_api_key: str | RerankingHeadersProvider | UnsetType = _UNSET,
        timeout_options: TimeoutOptions | UnsetType = _UNSET,
        serdes_options: SerdesOptions | UnsetType = _UNSET,
        data_api_url_options: DataAPIURLOptions | UnsetType = _UNSET,
        dev_ops_api_url_options: DevOpsAPIURLOptions | UnsetType = _UNSET,
    ) -> None:
        # Special conversions and type coercions occur here
        self.environment = _UNSET
        self.callers = callers
        self.database_additional_headers = database_additional_headers
        self.admin_additional_headers = admin_additional_headers
        self.redacted_header_names = (
            _UNSET
            if isinstance(redacted_header_names, UnsetType)
            else set(redacted_header_names)
        )
        self.token = coerce_possible_token_provider(token)
        self.embedding_api_key = coerce_possible_embedding_headers_provider(
            embedding_api_key,
        )
        self.reranking_api_key = coerce_possible_reranking_headers_provider(
            reranking_api_key,
        )
        self.timeout_options = timeout_options
        self.serdes_options = serdes_options
        self.data_api_url_options = data_api_url_options
        self.dev_ops_api_url_options = dev_ops_api_url_options

    def __repr__(self) -> str:
        # special items
        _admin_additional_headers: dict[str, str | None] | UnsetType
        _redacted_header_names = (
            set()
            if isinstance(self.redacted_header_names, UnsetType)
            else self.redacted_header_names
        )
        if not isinstance(self.admin_additional_headers, UnsetType):
            _admin_additional_headers = {
                k: v if k not in _redacted_header_names else FIXED_SECRET_PLACEHOLDER
                for k, v in self.admin_additional_headers.items()
            }
        else:
            _admin_additional_headers = _UNSET
        _database_additional_headers: dict[str, str | None] | UnsetType
        if not isinstance(self.database_additional_headers, UnsetType):
            _database_additional_headers = {
                k: v if k not in _redacted_header_names else FIXED_SECRET_PLACEHOLDER
                for k, v in self.database_additional_headers.items()
            }
        else:
            _database_additional_headers = _UNSET
        _token_desc: str | None
        if not isinstance(self.token, UnsetType) and self.token:
            _token_desc = f"token={self.token}"
        else:
            _token_desc = None

        non_unset_pieces = [
            pc
            for pc in (
                None
                if isinstance(self.callers, UnsetType)
                else f"callers={self.callers}",
                None
                if isinstance(_database_additional_headers, UnsetType)
                else f"database_additional_headers={_database_additional_headers}",
                None
                if isinstance(_admin_additional_headers, UnsetType)
                else f"admin_additional_headers={_admin_additional_headers}",
                None
                if isinstance(self.redacted_header_names, UnsetType)
                else f"redacted_header_names={self.redacted_header_names}",
                _token_desc,
                None
                if isinstance(self.embedding_api_key, UnsetType)
                else f"embedding_api_key={self.embedding_api_key}",
                None
                if isinstance(self.reranking_api_key, UnsetType)
                else f"reranking_api_key={self.reranking_api_key}",
                None
                if isinstance(self.timeout_options, UnsetType)
                else f"timeout_options={self.timeout_options}",
                None
                if isinstance(self.serdes_options, UnsetType)
                else f"serdes_options={self.serdes_options}",
                None
                if isinstance(self.data_api_url_options, UnsetType)
                else f"data_api_url_options={self.data_api_url_options}",
                None
                if isinstance(self.dev_ops_api_url_options, UnsetType)
                else f"dev_ops_api_url_options={self.dev_ops_api_url_options}",
            )
            if pc is not None
        ]
        inner_desc = ", ".join(non_unset_pieces)
        return f"{self.__class__.__name__}({inner_desc})"


@dataclass
class FullAPIOptions(APIOptions):
    """
    This class represents all settings that can be configured for how astrapy
    interacts with the API (both Data API and DevOps API). Each object in the
    abstraction hierarchy (DataAPIClient, Database, Table, Collection, ...)
    has a full set of these options that determine how it behaves when performing
    actions toward the API.

    This is the "full" version of the class, with the guarantee that all of its members
    have defined values. As such, this is what classes such as DataAPIClient, Database,
    Table, Collection and so on have as their `.api_options` attribute -- as opposed
    to the (non-full) `APIOptions` counterpart class: the latter admits "unset"
    attributes and is used to override specific settings. Please refer to the
    documentation for the `APIOptions` class for more details on how to customize
    the objects' behavior.

    Attributes:
        environment: an identifier for the environment for the Data API. This can
            describe an Astra DB environment (such as the default of "prod"), or
            a self-deployed setup (such as "dse" or "hcd"). This setting cannot be
            overridden through customization: it can only be provided when creating
            the DataAPIClient top object in the abstraction hierarchy.
        callers: an iterable of "caller identities" to be used in identifying the
            caller, through the User-Agent header, when issuing requests to the
            Data API. Each caller identity is a `(name, version)` 2-item tuple whose
            elements can be strings or None.
        database_additional_headers: free-form dictionary of additional headers to
            employ when issuing requests to the Data API from Database, Table and
            Collection classes. Passing a key with a value of None means that a certain
            header is suppressed when issuing the request.
        admin_additional_headers: free-form dictionary of additional headers to
            employ when issuing requests to both the Data API and the DevOps API
            from AstraDBAdmin, AstraDBDatabaseAdmin and DataAPIDatabaseAdmin classes.
            Passing a key with a value of None means that a certain header is
            suppressed when issuing the request.
        redacted_header_names: A set of (case-insensitive) strings denoting the headers
            that contain secrets, thus are to be masked when logging request details.
        token: an instance of TokenProvider to provide authentication to requests.
            Passing a string, or None, to this constructor parameter will get it
            automatically converted into the appropriate TokenProvider object.
            Depending on the target (Data API or DevOps API), this one attribute
            is encoded in the request appropriately.
        embedding_api_key: an instance of EmbeddingHeadersProvider should it be needed
            for vectorize-related data operations (used by Tables and Collections).
            Passing a string, or None, to this constructor parameter will get it
            automatically converted into the appropriate EmbeddingHeadersProvider
            object.
        reranking_api_key: an instance of RerankingHeadersProvider should it be needed
            for reranking-related data operations (used by Tables and Collections).
            Passing a string, or None, to this constructor parameter will get it
            automatically converted into the appropriate RerankingHeadersProvider
            object.
        timeout_options: an instance of `TimeoutOptions` (see) to control the timeout
            behavior for the various kinds of operations involving the Data/DevOps API.
        serdes_options: an instance of `SerdesOptions` (see) to customize the
            serializing/deserializing behavior related to writing to and reading from
            tables and collections.
        data_api_url_options: an instance of `DataAPIURLOptions` (see) to customize
            the full URL used to reach the Data API (customizing this setting
            is rarely needed).
        dev_ops_api_url_options: an instance of `DevOpsAPIURLOptions` (see) to
            customize the URL used to reach the DevOps API (customizing this setting
            is rarely needed; relevant only for Astra DB environments).
    """

    environment: str
    callers: Sequence[CallerType]
    database_additional_headers: dict[str, str | None]
    admin_additional_headers: dict[str, str | None]
    redacted_header_names: set[str]
    token: TokenProvider
    embedding_api_key: EmbeddingHeadersProvider
    reranking_api_key: RerankingHeadersProvider

    timeout_options: FullTimeoutOptions
    serdes_options: FullSerdesOptions
    data_api_url_options: FullDataAPIURLOptions
    dev_ops_api_url_options: FullDevOpsAPIURLOptions

    def __init__(
        self,
        *,
        environment: str,
        callers: Sequence[CallerType],
        database_additional_headers: dict[str, str | None],
        admin_additional_headers: dict[str, str | None],
        redacted_header_names: set[str],
        token: str | TokenProvider,
        embedding_api_key: str | EmbeddingHeadersProvider,
        reranking_api_key: str | RerankingHeadersProvider,
        timeout_options: FullTimeoutOptions,
        serdes_options: FullSerdesOptions,
        data_api_url_options: FullDataAPIURLOptions,
        dev_ops_api_url_options: FullDevOpsAPIURLOptions,
    ) -> None:
        APIOptions.__init__(
            self,
            callers=callers,
            database_additional_headers=database_additional_headers,
            admin_additional_headers=admin_additional_headers,
            redacted_header_names=redacted_header_names,
            token=token,
            embedding_api_key=embedding_api_key,
            reranking_api_key=reranking_api_key,
            timeout_options=timeout_options,
            serdes_options=serdes_options,
            data_api_url_options=data_api_url_options,
            dev_ops_api_url_options=dev_ops_api_url_options,
        )
        self.environment = environment

    def __repr__(self) -> str:
        # special items
        _token_desc: str | None
        if self.token:
            _token_desc = f"token={self.token}"
        else:
            _token_desc = None

        non_unset_pieces = [
            pc
            for pc in (
                None
                if self.environment == Environment.PROD
                else f"environment={self.environment}",
                _token_desc,
                f"embedding_api_key={self.embedding_api_key}"
                if self.embedding_api_key
                else None,
                f"reranking_api_key={self.reranking_api_key}"
                if self.reranking_api_key
                else None,
                "...",
            )
            if pc is not None
        ]
        inner_desc = ", ".join(non_unset_pieces)
        return f"{self.__class__.__name__}({inner_desc})"

    def with_override(self, other: APIOptions | None | UnsetType) -> FullAPIOptions:
        """
        Given an "overriding" set of options, possibly not defined in all its
        attributes, apply the override logic and return a new full options object.

        The override logic acts hierarchically, so as to deal with attributes that
        are, in turn, options object of one type of another.

        The override logic is such that defined attributes completely replace the
        pre-existing ones, except for the case of `database_additional_headers`,
        `admin_additional_headers` and `redacted_header_names`, in which cases
        merging takes place.

        Args:
            other: a not-necessarily-fully-specified options object. All its defined
                settings take precedence.
        """

        if isinstance(other, UnsetType) or other is None:
            return self

        database_additional_headers: dict[str, str | None]
        admin_additional_headers: dict[str, str | None]
        redacted_header_names: set[str]

        timeout_options: FullTimeoutOptions
        serdes_options: FullSerdesOptions
        data_api_url_options: FullDataAPIURLOptions
        dev_ops_api_url_options: FullDevOpsAPIURLOptions

        if isinstance(other.database_additional_headers, UnsetType):
            database_additional_headers = self.database_additional_headers
        else:
            database_additional_headers = {
                **self.database_additional_headers,
                **other.database_additional_headers,
            }
        if isinstance(other.admin_additional_headers, UnsetType):
            admin_additional_headers = self.admin_additional_headers
        else:
            admin_additional_headers = {
                **self.admin_additional_headers,
                **other.admin_additional_headers,
            }
        if isinstance(other.redacted_header_names, UnsetType):
            redacted_header_names = self.redacted_header_names
        else:
            redacted_header_names = (
                self.redacted_header_names | other.redacted_header_names
            )

        if isinstance(other.timeout_options, TimeoutOptions):
            timeout_options = self.timeout_options.with_override(other.timeout_options)
        else:
            timeout_options = self.timeout_options
        if isinstance(other.serdes_options, SerdesOptions):
            serdes_options = self.serdes_options.with_override(other.serdes_options)
        else:
            serdes_options = self.serdes_options
        if isinstance(other.data_api_url_options, DataAPIURLOptions):
            data_api_url_options = self.data_api_url_options.with_override(
                other.data_api_url_options
            )
        else:
            data_api_url_options = self.data_api_url_options
        if isinstance(other.dev_ops_api_url_options, DevOpsAPIURLOptions):
            dev_ops_api_url_options = self.dev_ops_api_url_options.with_override(
                other.dev_ops_api_url_options
            )
        else:
            dev_ops_api_url_options = self.dev_ops_api_url_options

        return FullAPIOptions(
            environment=(
                other.environment
                if not isinstance(other.environment, UnsetType)
                else self.environment
            ),
            callers=(
                other.callers
                if not isinstance(other.callers, UnsetType)
                else self.callers
            ),
            database_additional_headers=database_additional_headers,
            admin_additional_headers=admin_additional_headers,
            redacted_header_names=redacted_header_names,
            token=other.token if not isinstance(other.token, UnsetType) else self.token,
            embedding_api_key=(
                other.embedding_api_key
                if not isinstance(other.embedding_api_key, UnsetType)
                else self.embedding_api_key
            ),
            reranking_api_key=(
                other.reranking_api_key
                if not isinstance(other.reranking_api_key, UnsetType)
                else self.reranking_api_key
            ),
            timeout_options=timeout_options,
            serdes_options=serdes_options,
            data_api_url_options=data_api_url_options,
            dev_ops_api_url_options=dev_ops_api_url_options,
        )


defaultTimeoutOptions = FullTimeoutOptions(
    request_timeout_ms=DEFAULT_REQUEST_TIMEOUT_MS,
    general_method_timeout_ms=DEFAULT_GENERAL_METHOD_TIMEOUT_MS,
    collection_admin_timeout_ms=DEFAULT_COLLECTION_ADMIN_TIMEOUT_MS,
    table_admin_timeout_ms=DEFAULT_TABLE_ADMIN_TIMEOUT_MS,
    database_admin_timeout_ms=DEFAULT_DATABASE_ADMIN_TIMEOUT_MS,
    keyspace_admin_timeout_ms=DEFAULT_KEYSPACE_ADMIN_TIMEOUT_MS,
)
defaultSerdesOptions = FullSerdesOptions(
    binary_encode_vectors=DEFAULT_BINARY_ENCODE_VECTORS,
    custom_datatypes_in_reading=DEFAULT_CUSTOM_DATATYPES_IN_READING,
    unroll_iterables_to_lists=DEFAULT_UNROLL_ITERABLES_TO_LISTS,
    use_decimals_in_collections=DEFAULT_USE_DECIMALS_IN_COLLECTIONS,
    encode_maps_as_lists_in_tables=DEFAULT_ENCODE_MAPS_AS_LISTS_IN_TABLES,
    accept_naive_datetimes=DEFAULT_ACCEPT_NAIVE_DATETIMES,
    datetime_tzinfo=DEFAULT_DATETIME_TZINFO,
)


def defaultAPIOptions(environment: str) -> FullAPIOptions:
    """
    Return the default APIOptions object for a given environment,
    based on 'grand defaults' hardcoded in astrapy.
    """

    defaultDataAPIURLOptions = FullDataAPIURLOptions(
        api_path=API_PATH_ENV_MAP[environment],
        api_version=API_VERSION_ENV_MAP[environment],
    )
    defaultDevOpsAPIURLOptions: FullDevOpsAPIURLOptions
    if environment in Environment.astra_db_values:
        defaultDevOpsAPIURLOptions = FullDevOpsAPIURLOptions(
            dev_ops_url=DEV_OPS_URL_ENV_MAP[environment],
            dev_ops_api_version=DEV_OPS_VERSION_ENV_MAP[environment],
        )
    else:
        defaultDevOpsAPIURLOptions = FullDevOpsAPIURLOptions(
            dev_ops_url="never used",
            dev_ops_api_version=None,
        )
    return FullAPIOptions(
        environment=environment,
        callers=[],
        database_additional_headers={},
        admin_additional_headers={},
        redacted_header_names=set(),
        token=StaticTokenProvider(None),
        embedding_api_key=EmbeddingAPIKeyHeaderProvider(None),
        reranking_api_key=RerankingAPIKeyHeaderProvider(None),
        timeout_options=defaultTimeoutOptions,
        serdes_options=defaultSerdesOptions,
        data_api_url_options=defaultDataAPIURLOptions,
        dev_ops_api_url_options=defaultDevOpsAPIURLOptions,
    )



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import datetime
from typing import Any, Dict, cast

from astrapy.constants import DefaultDocumentType
from astrapy.data.utils.extended_json_converters import (
    convert_ejson_binary_object_to_bytes,
    convert_ejson_date_object_to_apitimestamp,
    convert_ejson_date_object_to_datetime,
    convert_ejson_objectid_object_to_objectid,
    convert_ejson_uuid_object_to_uuid,
    convert_to_ejson_apitimestamp_object,
    convert_to_ejson_bytes,
    convert_to_ejson_date_object,
    convert_to_ejson_objectid_object,
    convert_to_ejson_uuid_object,
)
from astrapy.data.utils.vector_coercion import (
    convert_vector_to_floats,
    ensure_unrolled_if_iterable,
    is_list_of_floats,
)
from astrapy.data_types import DataAPITimestamp, DataAPIVector
from astrapy.ids import UUID, ObjectId
from astrapy.settings.error_messages import CANNOT_ENCODE_NAIVE_DATETIME_ERROR_MESSAGE
from astrapy.utils.api_options import FullSerdesOptions

FIND_AND_RERANK_VECTOR_FLOAT_PATH = [
    "status",
    "documentResponses",
    "",
    "scores",
    "$vector",
]


def preprocess_collection_payload_value(
    path: list[str], value: Any, options: FullSerdesOptions
) -> Any:
    """
    The path helps determining special treatments
    """

    # vector-related pre-processing and coercion
    _value = value
    # is this value in the place for vectors?
    if path[-1:] == ["$vector"] and path[-2:] != ["projection", "$vector"]:
        # must coerce list-likes broadly, and is it the case to do it?
        if options.unroll_iterables_to_lists and not (
            is_list_of_floats(_value) or isinstance(_value, DataAPIVector)
        ):
            _value = convert_vector_to_floats(_value)
        # now _value is either a list or a DataAPIVector.
        # can/should it be binary-encoded?
        can_bin_encode = path[0] in {"insertOne", "insertMany"}
        # will it be bin-encoded?
        if isinstance(_value, DataAPIVector):
            # if I can, I will
            if can_bin_encode and options.binary_encode_vectors:
                return convert_to_ejson_bytes(_value.to_bytes())
            else:
                # back to a regular list
                return _value.data
        else:
            # this is a list. Encode if serdes options allow it
            if can_bin_encode and options.binary_encode_vectors:
                return convert_to_ejson_bytes(DataAPIVector(_value).to_bytes())
            else:
                return _value

    if options.unroll_iterables_to_lists:
        _value = ensure_unrolled_if_iterable(_value)
    if isinstance(_value, dict):
        return {
            k: preprocess_collection_payload_value(path + [k], v, options=options)
            for k, v in _value.items()
        }
    elif isinstance(_value, list):
        return [
            preprocess_collection_payload_value(path + [""], list_item, options=options)
            for list_item in _value
        ]
    elif isinstance(_value, datetime.datetime):
        if _value.utcoffset() is None and not options.accept_naive_datetimes:
            raise ValueError(CANNOT_ENCODE_NAIVE_DATETIME_ERROR_MESSAGE)
        return convert_to_ejson_date_object(_value)
    elif isinstance(_value, datetime.date):
        # Note: since 'datetime' subclasses 'date', this must come after the previous.
        return convert_to_ejson_date_object(_value)
    elif isinstance(_value, bytes):
        return convert_to_ejson_bytes(_value)
    elif isinstance(_value, UUID):
        return convert_to_ejson_uuid_object(_value)
    elif isinstance(_value, ObjectId):
        return convert_to_ejson_objectid_object(_value)
    elif isinstance(_value, DataAPITimestamp):
        return convert_to_ejson_apitimestamp_object(_value)
    else:
        return _value


def preprocess_collection_payload(
    payload: dict[str, Any] | None, options: FullSerdesOptions
) -> dict[str, Any] | None:
    """
    Normalize a payload for API calls.
    This includes e.g. ensuring values for "$vector" key are properly coerced.

    Args:
        payload (dict[str, Any]): A dict expressing a payload for an API call

    Returns:
        dict[str, Any]: a payload dict, pre-processed, ready for HTTP requests.
    """

    if payload:
        return cast(
            Dict[str, Any],
            preprocess_collection_payload_value([], payload, options=options),
        )
    else:
        return payload


def postprocess_collection_response_value(
    path: list[str], value: Any, options: FullSerdesOptions
) -> Any:
    """
    The path helps determining special treatments
    """

    # for reads, (almost) everywhere there's a $vector it can be treated as such and reconverted
    if path[-1:] == ["$vector"] and path != FIND_AND_RERANK_VECTOR_FLOAT_PATH:
        # custom faster handling for the $vector path:
        if isinstance(value, list):
            if options.custom_datatypes_in_reading:
                return DataAPIVector(value)
            else:
                return value
        elif isinstance(value, dict):
            _bytes = convert_ejson_binary_object_to_bytes(value)
            if options.custom_datatypes_in_reading:
                return DataAPIVector.from_bytes(_bytes)
            else:
                return DataAPIVector.from_bytes(_bytes).data
        else:
            raise ValueError(
                f"Response parsing failed at path '{path}': unexpected data type "
                f"found under $vector: {type(value)}"
            )
    if isinstance(value, dict):
        value_keys = set(value.keys())
        if value_keys == {"$date"}:
            # this is `{"$date": 123456}`.
            # Restore to the appropriate APIOptions-required object
            if options.custom_datatypes_in_reading:
                return convert_ejson_date_object_to_apitimestamp(value)
            else:
                return convert_ejson_date_object_to_datetime(
                    value, tz=options.datetime_tzinfo
                )
        elif value_keys == {"$uuid"}:
            # this is `{"$uuid": "abc123..."}`, restore to UUID
            return convert_ejson_uuid_object_to_uuid(value)
        elif value_keys == {"$objectId"}:
            # this is `{"$objectId": "123abc..."}`, restore to ObjectId
            return convert_ejson_objectid_object_to_objectid(value)
        elif value_keys == {"$binary"}:
            # this is `{"$binary": "xyz=="}`, restore to `bytes`
            return convert_ejson_binary_object_to_bytes(value)
        else:
            return {
                k: postprocess_collection_response_value(path + [k], v, options=options)
                for k, v in value.items()
            }
    elif isinstance(value, list):
        return [
            postprocess_collection_response_value(
                path + [""], list_item, options=options
            )
            for list_item in value
        ]
    else:
        return value


def postprocess_collection_response(
    response: DefaultDocumentType, options: FullSerdesOptions
) -> DefaultDocumentType:
    """
    Process a dictionary just returned from the API.
    This is the place where e.g. `{"$date": 123}` is
    converted back into a datetime object.
    """
    return cast(
        DefaultDocumentType,
        postprocess_collection_response_value([], response, options=options),
    )



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any

from astrapy.data.info.vectorize import VectorServiceOptions
from astrapy.data.utils.table_types import (
    ColumnType,
    TableKeyValuedColumnType,
    TableUnsupportedColumnType,
    TableValuedColumnType,
    TableVectorColumnType,
)
from astrapy.utils.parsing import _warn_residual_keys


@dataclass
class TableAPISupportDescriptor:
    """
    Represents the additional support information returned by the Data API when
    describing columns of a table. Some columns indeed require a detailed description
    of what operations are supported on them - this includes, but is not limited to,
    columns created by means other than the Data API (e.g. CQL direct interaction
    with the database).

    When the Data API reports these columns (in listing the tables and their metadata),
    it provides the information marshaled in this object to detail which level
    of support the column has (for instance, it can be a partial support whereby the
    column is readable by the API but not writable).

    Attributes:
        cql_definition: a free-form string containing the CQL definition for the column.
        create_table: whether a column of this nature can be used in API table creation.
        insert: whether a column of this nature can be written through the API.
        filter: whether a column of this nature can be used for filtering with API find.
        read: whether a column of this nature can be read through the API.
    """

    cql_definition: str
    create_table: bool
    insert: bool
    filter: bool
    read: bool

    def __repr__(self) -> str:
        desc = ", ".join(
            [
                f'"{self.cql_definition}"',
                f"create_table={self.create_table}",
                f"insert={self.insert}",
                f"filter={self.filter}",
                f"read={self.read}",
            ]
        )
        return f"{self.__class__.__name__}({desc})"

    def as_dict(self) -> dict[str, Any]:
        """Recast this object into a dictionary."""

        return {
            "cqlDefinition": self.cql_definition,
            "createTable": self.create_table,
            "insert": self.insert,
            "filter": self.filter,
            "read": self.read,
        }

    @classmethod
    def _from_dict(cls, raw_dict: dict[str, Any]) -> TableAPISupportDescriptor:
        """
        Create an instance of TableAPISupportDescriptor from a dictionary
        such as one from the Data API.
        """

        _warn_residual_keys(
            cls,
            raw_dict,
            {"cqlDefinition", "createTable", "insert", "filter", "read"},
        )
        return TableAPISupportDescriptor(
            cql_definition=raw_dict["cqlDefinition"],
            create_table=raw_dict["createTable"],
            insert=raw_dict["insert"],
            filter=raw_dict["filter"],
            read=raw_dict["read"],
        )


@dataclass
class TableColumnTypeDescriptor(ABC):
    """
    Represents and describes a column in a Table, with its type and any
    additional property.

    This is an abstract class, whose concrete implementation are the various
    kinds of column descriptors such as `TableScalarColumnTypeDescriptor`,
    `TableVectorColumnTypeDescriptor`, `TableValuedColumnTypeDescriptor`, and so on.

    Attributes:
        column_type: an instance of one of the various column-type classes, according
            to the type of the column. In other words, each subclass of
            `TableColumnTypeDescriptor` has an appropriate object as its
            `column_type` attributes.
            For example the `column_type` of `TableValuedColumnTypeDescriptor`
            is a `TableValuedColumnType`.
        api_support: a `TableAPISupportDescriptor` object giving more details.
    """

    column_type: (
        ColumnType
        | TableValuedColumnType
        | TableKeyValuedColumnType
        | TableVectorColumnType
        | TableUnsupportedColumnType
    )
    api_support: TableAPISupportDescriptor | None

    @abstractmethod
    def as_dict(self) -> dict[str, Any]: ...

    @classmethod
    def _from_dict(cls, raw_dict: dict[str, Any]) -> TableColumnTypeDescriptor:
        """
        Create an instance of TableColumnTypeDescriptor from a dictionary
        such as one from the Data API.

        This method switches to the proper subclass depending on the input.
        """

        if "keyType" in raw_dict:
            return TableKeyValuedColumnTypeDescriptor._from_dict(raw_dict)
        elif "valueType" in raw_dict:
            return TableValuedColumnTypeDescriptor._from_dict(raw_dict)
        elif raw_dict["type"] == "vector":
            return TableVectorColumnTypeDescriptor._from_dict(raw_dict)
        elif raw_dict["type"] == "UNSUPPORTED":
            return TableUnsupportedColumnTypeDescriptor._from_dict(raw_dict)
        else:
            return TableScalarColumnTypeDescriptor._from_dict(raw_dict)

    @classmethod
    def coerce(
        cls, raw_input: TableColumnTypeDescriptor | dict[str, Any] | str
    ) -> TableColumnTypeDescriptor:
        """
        Normalize the input, whether an object already or a plain dictionary
        of the right structure, into a TableColumnTypeDescriptor.
        """

        if isinstance(raw_input, TableColumnTypeDescriptor):
            return raw_input
        elif isinstance(raw_input, str):
            return cls._from_dict({"type": raw_input})
        else:
            return cls._from_dict(raw_input)


@dataclass
class TableScalarColumnTypeDescriptor(TableColumnTypeDescriptor):
    """
    Represents and describes a column in a Table, of scalar type, i.e. which contains
    a single simple value.

    Attributes:
        column_type: a `ColumnType` value. When creating the object,
            simple strings such as "TEXT" or "UUID" are also accepted.
        api_support: a `TableAPISupportDescriptor` object giving more details.
    """

    column_type: ColumnType

    def __init__(
        self,
        column_type: str | ColumnType,
        api_support: TableAPISupportDescriptor | None = None,
    ) -> None:
        super().__init__(
            column_type=ColumnType.coerce(column_type),
            api_support=api_support,
        )

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.column_type.value})"

    def as_dict(self) -> dict[str, Any]:
        """Recast this object into a dictionary."""

        return {
            k: v
            for k, v in {
                "type": self.column_type.value,
                "apiSupport": self.api_support.as_dict() if self.api_support else None,
            }.items()
            if v is not None
        }

    @classmethod
    def _from_dict(cls, raw_dict: dict[str, Any]) -> TableScalarColumnTypeDescriptor:
        """
        Create an instance of TableScalarColumnTypeDescriptor from a dictionary
        such as one from the Data API.
        """

        _warn_residual_keys(cls, raw_dict, {"type", "apiSupport"})
        return TableScalarColumnTypeDescriptor(
            column_type=raw_dict["type"],
            api_support=TableAPISupportDescriptor._from_dict(raw_dict["apiSupport"])
            if raw_dict.get("apiSupport")
            else None,
        )


@dataclass
class TableVectorColumnTypeDescriptor(TableColumnTypeDescriptor):
    """
    Represents and describes a column in a Table, of vector type, i.e. which contains
    a list of `dimension` floats that is treated specially as a "vector".

    Attributes:
        column_type: a `TableVectorColumnType` value. This can be omitted when
            creating the object. It only ever assumes the "VECTOR" value.
        dimension: an integer, the number of components (numbers) in the vectors.
            This can be left unspecified in some cases of vectorize-enabled columns.
        service: an optional `VectorServiceOptions` object defining the vectorize
            settings (i.e. server-side embedding computation) for the column.
        api_support: a `TableAPISupportDescriptor` object giving more details.
    """

    column_type: TableVectorColumnType
    dimension: int | None
    service: VectorServiceOptions | None

    def __init__(
        self,
        *,
        column_type: str | TableVectorColumnType = TableVectorColumnType.VECTOR,
        dimension: int | None,
        service: VectorServiceOptions | None = None,
        api_support: TableAPISupportDescriptor | None = None,
    ) -> None:
        self.dimension = dimension
        self.service = service
        super().__init__(
            column_type=TableVectorColumnType.coerce(column_type),
            api_support=api_support,
        )

    def __repr__(self) -> str:
        not_null_pieces = [
            pc
            for pc in [
                f"dimension={self.dimension}" if self.dimension is not None else None,
                None if self.service is None else f"service={self.service}",
            ]
            if pc is not None
        ]
        inner_desc = ", ".join(not_null_pieces)

        return f"{self.__class__.__name__}({self.column_type.value}[{inner_desc}])"

    def as_dict(self) -> dict[str, Any]:
        """Recast this object into a dictionary."""

        return {
            k: v
            for k, v in {
                "type": self.column_type.value,
                "dimension": self.dimension,
                "service": None if self.service is None else self.service.as_dict(),
                "apiSupport": self.api_support.as_dict() if self.api_support else None,
            }.items()
            if v is not None
        }

    @classmethod
    def _from_dict(cls, raw_dict: dict[str, Any]) -> TableVectorColumnTypeDescriptor:
        """
        Create an instance of TableVectorColumnTypeDescriptor from a dictionary
        such as one from the Data API.
        """

        _warn_residual_keys(
            cls,
            raw_dict,
            {"type", "dimension", "service", "apiSupport"},
        )
        return TableVectorColumnTypeDescriptor(
            column_type=raw_dict["type"],
            dimension=raw_dict.get("dimension"),
            service=VectorServiceOptions.coerce(raw_dict.get("service")),
            api_support=TableAPISupportDescriptor._from_dict(raw_dict["apiSupport"])
            if raw_dict.get("apiSupport")
            else None,
        )


@dataclass
class TableValuedColumnTypeDescriptor(TableColumnTypeDescriptor):
    """
    Represents and describes a column in a Table, of a 'valued' type that stores
    multiple values. This means either a list or a set of homogeneous items.

    Attributes:
        column_type: an instance of `TableValuedColumnType`. When creating the
            object, simple strings such as "list" or "set" are also accepted.
        value_type: the type of the individual items stored in the column.
            This is a `ColumnType`, but when creating the object,
            strings such as "TEXT" or "UUID" are also accepted.
        api_support: a `TableAPISupportDescriptor` object giving more details.
    """

    column_type: TableValuedColumnType
    value_type: ColumnType

    def __init__(
        self,
        *,
        column_type: str | TableValuedColumnType,
        value_type: str | ColumnType,
        api_support: TableAPISupportDescriptor | None = None,
    ) -> None:
        self.value_type = ColumnType.coerce(value_type)
        super().__init__(
            column_type=TableValuedColumnType.coerce(column_type),
            api_support=api_support,
        )

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}({self.column_type.value}"
            f"<{self.value_type.value}>)"
        )

    def as_dict(self) -> dict[str, Any]:
        """Recast this object into a dictionary."""

        return {
            k: v
            for k, v in {
                "type": self.column_type.value,
                "valueType": self.value_type.value,
                "apiSupport": self.api_support.as_dict() if self.api_support else None,
            }.items()
            if v is not None
        }

    @classmethod
    def _from_dict(cls, raw_dict: dict[str, Any]) -> TableValuedColumnTypeDescriptor:
        """
        Create an instance of TableValuedColumnTypeDescriptor from a dictionary
        such as one from the Data API.
        """

        _warn_residual_keys(cls, raw_dict, {"type", "valueType", "apiSupport"})
        return TableValuedColumnTypeDescriptor(
            column_type=raw_dict["type"],
            value_type=raw_dict["valueType"],
            api_support=TableAPISupportDescriptor._from_dict(raw_dict["apiSupport"])
            if raw_dict.get("apiSupport")
            else None,
        )


@dataclass
class TableKeyValuedColumnTypeDescriptor(TableColumnTypeDescriptor):
    """
    Represents and describes a column in a Table, of a 'key-value' type, that stores
    an associative map (essentially a dict) between keys of a certain scalar type and
    values of a certain scalar type. The only such kind of column is a "map".

    Attributes:
        column_type: an instance of `TableKeyValuedColumnType`. When creating the
            object, this can be omitted as it only ever assumes the "MAP" value.
        key_type: the type of the individual keys in the map column.
            This is a `ColumnType`, but when creating the object,
            strings such as "TEXT" or "UUID" are also accepted.
        value_type: the type of the individual values stored in the map for a single key.
            This is a `ColumnType`, but when creating the object,
            strings such as "TEXT" or "UUID" are also accepted.
        api_support: a `TableAPISupportDescriptor` object giving more details.
    """

    column_type: TableKeyValuedColumnType
    key_type: ColumnType
    value_type: ColumnType

    def __init__(
        self,
        *,
        value_type: str | ColumnType,
        key_type: str | ColumnType,
        column_type: str | TableKeyValuedColumnType = TableKeyValuedColumnType.MAP,
        api_support: TableAPISupportDescriptor | None = None,
    ) -> None:
        self.key_type = ColumnType.coerce(key_type)
        self.value_type = ColumnType.coerce(value_type)
        super().__init__(
            column_type=TableKeyValuedColumnType.coerce(column_type),
            api_support=api_support,
        )

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}({self.column_type.value}"
            f"<{self.key_type.value},{self.value_type.value}>)"
        )

    def as_dict(self) -> dict[str, Any]:
        """Recast this object into a dictionary."""

        return {
            k: v
            for k, v in {
                "type": self.column_type.value,
                "keyType": self.key_type.value,
                "valueType": self.value_type.value,
                "apiSupport": self.api_support.as_dict() if self.api_support else None,
            }.items()
            if v is not None
        }

    @classmethod
    def _from_dict(cls, raw_dict: dict[str, Any]) -> TableKeyValuedColumnTypeDescriptor:
        """
        Create an instance of TableKeyValuedColumnTypeDescriptor from a dictionary
        such as one from the Data API.
        """

        _warn_residual_keys(
            cls, raw_dict, {"type", "keyType", "valueType", "apiSupport"}
        )
        return TableKeyValuedColumnTypeDescriptor(
            column_type=raw_dict["type"],
            key_type=raw_dict["keyType"],
            value_type=raw_dict["valueType"],
            api_support=TableAPISupportDescriptor._from_dict(raw_dict["apiSupport"])
            if raw_dict.get("apiSupport")
            else None,
        )


@dataclass
class TableUnsupportedColumnTypeDescriptor(TableColumnTypeDescriptor):
    """
    Represents and describes a column in a Table, of unsupported type.

    Note that this column type descriptor cannot be used in table creation,
    rather it can only be returned when listing the tables or getting their
    metadata by the API.

    Attributes:
        column_type: an instance of `TableUnsupportedColumnType`.
        api_support: a `TableAPISupportDescriptor` object giving more details.

    This class has no `coerce` method, since it is always only found in API responses.
    """

    column_type: TableUnsupportedColumnType
    api_support: TableAPISupportDescriptor

    def __init__(
        self,
        *,
        column_type: TableUnsupportedColumnType | str,
        api_support: TableAPISupportDescriptor,
    ) -> None:
        super().__init__(
            column_type=TableUnsupportedColumnType.coerce(column_type),
            api_support=api_support,
        )

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.api_support.cql_definition})"

    def as_dict(self) -> dict[str, Any]:
        """Recast this object into a dictionary."""

        return {
            k: v
            for k, v in {
                "type": self.column_type.value,
                "apiSupport": self.api_support.as_dict(),
            }.items()
            if v is not None
        }

    @classmethod
    def _from_dict(
        cls, raw_dict: dict[str, Any]
    ) -> TableUnsupportedColumnTypeDescriptor:
        """
        Create an instance of TableUnsupportedColumnTypeDescriptor from a dictionary
        such as one from the Data API.
        """

        _warn_residual_keys(cls, raw_dict, {"type", "apiSupport"})
        return TableUnsupportedColumnTypeDescriptor(
            column_type=raw_dict["type"],
            api_support=TableAPISupportDescriptor._from_dict(raw_dict["apiSupport"]),
        )


@dataclass
class TablePrimaryKeyDescriptor:
    """
    Represents the part of a table definition that describes the primary key.

    Attributes:
        partition_by: a list of column names forming the partition key, i.e.
            the portion of primary key that determines physical grouping and storage
            of rows on the database. Rows with the same values for the partition_by
            columns are guaranteed to be stored next to each other. This list
            cannot be empty.
        partition_sort: this defines how rows are to be sorted within a partition.
            It is a dictionary that specifies, for each column of the primary key
            not in the `partition_by` field, whether the sorting is ascending
            or descending (see the values in the `SortMode` constant).
            The sorting within a partition considers all columns in this dictionary,
            in a hierarchical way: hence, ordering in this dictionary is relevant.
    """

    partition_by: list[str]
    partition_sort: dict[str, int]

    def __repr__(self) -> str:
        partition_key_block = ",".join(self.partition_by)
        clustering_block = ",".join(
            f"{clu_col_name}:{'a' if clu_col_sort > 0 else 'd'}"
            for clu_col_name, clu_col_sort in self.partition_sort.items()
        )
        pk_block = f"({partition_key_block}){clustering_block}"
        return f"{self.__class__.__name__}[{pk_block}]"

    def as_dict(self) -> dict[str, Any]:
        """Recast this object into a dictionary."""

        return {
            k: v
            for k, v in {
                "partitionBy": self.partition_by,
                "partitionSort": dict(self.partition_sort.items()),
            }.items()
            if v is not None
        }

    @classmethod
    def _from_dict(cls, raw_dict: dict[str, Any]) -> TablePrimaryKeyDescriptor:
        """
        Create an instance of TablePrimaryKeyDescriptor from a dictionary
        such as one from the Data API.
        """

        _warn_residual_keys(cls, raw_dict, {"partitionBy", "partitionSort"})
        return TablePrimaryKeyDescriptor(
            partition_by=raw_dict["partitionBy"],
            partition_sort=raw_dict["partitionSort"],
        )

    @classmethod
    def coerce(
        cls, raw_input: TablePrimaryKeyDescriptor | dict[str, Any] | str
    ) -> TablePrimaryKeyDescriptor:
        """
        Normalize the input, whether an object already or a plain dictionary
        of the right structure, into a TablePrimaryKeyDescriptor.
        """

        if isinstance(raw_input, TablePrimaryKeyDescriptor):
            return raw_input
        elif isinstance(raw_input, str):
            return cls._from_dict({"partitionBy": [raw_input], "partitionSort": {}})
        else:
            return cls._from_dict(raw_input)



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import base64
import datetime
import time

from astrapy.data_types import DataAPITimestamp
from astrapy.ids import UUID, ObjectId


def convert_to_ejson_date_object(
    date_value: datetime.date | datetime.datetime,
) -> dict[str, int]:
    if isinstance(date_value, datetime.datetime):
        return {"$date": int(date_value.timestamp() * 1000)}
    return {"$date": int(time.mktime(date_value.timetuple()) * 1000)}


def convert_to_ejson_apitimestamp_object(
    date_value: DataAPITimestamp,
) -> dict[str, int]:
    return {"$date": date_value.timestamp_ms}


def convert_to_ejson_bytes(bytes_value: bytes) -> dict[str, str]:
    return {"$binary": base64.b64encode(bytes_value).decode()}


def convert_to_ejson_uuid_object(uuid_value: UUID) -> dict[str, str]:
    return {"$uuid": str(uuid_value)}


def convert_to_ejson_objectid_object(objectid_value: ObjectId) -> dict[str, str]:
    return {"$objectId": str(objectid_value)}


def convert_ejson_date_object_to_datetime(
    date_object: dict[str, int], tz: datetime.timezone | None
) -> datetime.datetime:
    return datetime.datetime.fromtimestamp(date_object["$date"] / 1000.0, tz=tz)


def convert_ejson_date_object_to_apitimestamp(
    date_object: dict[str, int],
) -> DataAPITimestamp:
    return DataAPITimestamp(date_object["$date"])


def convert_ejson_binary_object_to_bytes(
    binary_object: dict[str, str],
) -> bytes:
    return base64.b64decode(binary_object["$binary"])


def convert_ejson_uuid_object_to_uuid(uuid_object: dict[str, str]) -> UUID:
    return UUID(uuid_object["$uuid"])


def convert_ejson_objectid_object_to_objectid(
    objectid_object: dict[str, str],
) -> ObjectId:
    return ObjectId(objectid_object["$objectId"])



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import math
from typing import Generic, Iterable, Iterator, Mapping, TypeVar

T = TypeVar("T")
U = TypeVar("U")


def _accumulate_pairs(
    destination: tuple[list[T], list[U]], source: Iterable[tuple[T, U]]
) -> tuple[list[T], list[U]]:
    _new_ks = list(destination[0])
    _new_vs = list(destination[1])
    for k, v in source:
        if k not in _new_ks:
            _new_ks.append(k)
            _new_vs.append(v)
    return (_new_ks, _new_vs)


class DataAPIMap(Generic[T, U], Mapping[T, U]):
    """
    An immutable 'map-like' class that preserves the order and can employ
    non-hashable keys (which must support __eq__). Not designed for performance.

    Despite internally preserving the order, equality between DataAPIMap instances
    (and with regular dicts) is independent of the order.
    """

    _keys: list[T]
    _values: list[U]

    def __init__(self, source: Iterable[tuple[T, U]] | dict[T, U] = []) -> None:
        if isinstance(source, dict):
            self._keys, self._values = _accumulate_pairs(
                ([], []),
                source.items(),
            )
        else:
            self._keys, self._values = _accumulate_pairs(
                ([], []),
                source,
            )

    def __getitem__(self, key: T) -> U:
        if isinstance(key, float) and math.isnan(key):
            for idx, k in enumerate(self._keys):
                if isinstance(k, float) and math.isnan(k):
                    return self._values[idx]
            raise KeyError(str(key))
        else:
            for idx, k in enumerate(self._keys):
                if k == key:
                    return self._values[idx]
            raise KeyError(str(key) + "//" + str(self._keys) + "//" + str(self._values))

    def __iter__(self) -> Iterator[T]:
        return iter(self._keys)

    def __len__(self) -> int:
        return len(self._keys)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, DataAPIMap):
            if len(self) == len(other):
                if all(o_k in self for o_k in other):
                    return all(other[k] == self[k] for k in self)
            return False
        try:
            dother = dict(other)  # type: ignore[call-overload]
            return all(
                [
                    len(dother) == len(self),
                    all(o_k in self for o_k in dother),
                    all(dother[k] == self[k] for k in self),
                ]
            )
        except KeyError:
            return False
        except TypeError:
            pass
        return NotImplemented

    def __repr__(self) -> str:
        _map_repr = ", ".join(
            f"({repr(k)}, {repr(v)})" for k, v in zip(self._keys, self._values)
        )
        return f"{self.__class__.__name__}([{_map_repr}])"

    def __str__(self) -> str:
        _map_repr = ", ".join(f"({k}, {v})" for k, v in zip(self._keys, self._values))
        return f"{_map_repr}"

    def __reduce__(self) -> tuple[type, tuple[Iterable[tuple[T, U]]]]:
        return self.__class__, (list(zip(self._keys, self._values)),)



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import datetime
import re
from dataclasses import dataclass
from typing import Any

from astrapy.utils.date_utils import (
    _validate_time,
)

TIME_PARSE_PATTERN = re.compile(r"^(\d+):(\d+)(:(\d+)(\.\d+)?)?$")
SECONDS_PARSE_PATTERN = re.compile(r"^:(\d+)(\.\d+)?$")

TIME_FORMAT_DESC = (
    "Times must be '<hour>:<minute>[:<second>[.<fractional-seconds>]]', with: "
    "hour in 0..23; minute in 0..59; second in 0..59; and optionally a decimal "
    "point followed by a fraction of second up to nanosecond precision (9 digits)."
)


@dataclass
class DataAPITime:
    """
    A value expressing a time, composed of hours, minutes, seconds and nanoseconds,
    suitable for working with the "time" table column type.

    This class is designed to losslessly express the time values the Data API
    supports, overcoming the precision limitation of Python's standard-library
    time (whose sub-second quantity only has microsecond precision).

    DataAPITime objects are meant to easily work in the context of the Data API,
    hence its conversion methods from/to a string assumed the particular format
    employed by the API.

    The class also offers conversion methods from/to the regular
    Python `datetime.time`; however these can entail a lossy conversion because
    of the missing support for nanoseconds by the latter.

    Args:
        hour: an integer in 0-23, the hour value of the time.
        minute: an integer in 0-59, the minute value of the time.
        second: an integer in 0-59, the second value of the time.
        nanosecond: an integer in 0-999999999, the nanosecond value of the time.

    Example:
        >>> from astrapy.data_types import DataAPITime
        >>>
        >>> DataAPITime(12, 34, 56)
        DataAPITime(12, 34, 56, 0)
        >>> t1 = DataAPITime(21, 43, 56, 789012345)
        >>> t1
        DataAPITime(21, 43, 56, 789012345)
        >>> t1.second
        56
    """

    hour: int
    minute: int
    second: int
    nanosecond: int

    def __init__(
        self, hour: int, minute: int = 0, second: int = 0, nanosecond: int = 0
    ):
        _fail_reason = _validate_time(
            hour=hour,
            minute=minute,
            second=second,
            nanosecond=nanosecond,
        )
        if _fail_reason:
            raise ValueError(f"Invalid time arguments: {_fail_reason}.")
        self.hour = hour
        self.minute = minute
        self.second = second
        self.nanosecond = nanosecond

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}({self.hour}, {self.minute}, "
            f"{self.second}, {self.nanosecond})"
        )

    def __hash__(self) -> int:
        return hash((self.hour, self.minute, self.second, self.nanosecond))

    def __str__(self) -> str:
        return self.to_string()

    def __reduce__(self) -> tuple[type, tuple[int, int, int, int]]:
        return self.__class__, (self.hour, self.minute, self.second, self.nanosecond)

    def __le__(self, other: Any) -> bool:
        if isinstance(other, DataAPITime):
            return self._to_tuple() <= other._to_tuple()
        elif isinstance(other, datetime.time):
            return self.__le__(DataAPITime.from_time(other))
        else:
            return NotImplemented

    def __lt__(self, other: Any) -> bool:
        if isinstance(other, DataAPITime):
            return self._to_tuple() < other._to_tuple()
        elif isinstance(other, datetime.time):
            return self.__lt__(DataAPITime.from_time(other))
        else:
            return NotImplemented

    def __ge__(self, other: Any) -> bool:
        if isinstance(other, DataAPITime):
            return self._to_tuple() >= other._to_tuple()
        elif isinstance(other, datetime.time):
            return self.__ge__(DataAPITime.from_time(other))
        else:
            return NotImplemented

    def __gt__(self, other: Any) -> bool:
        if isinstance(other, DataAPITime):
            return self._to_tuple() > other._to_tuple()
        elif isinstance(other, datetime.time):
            return self.__gt__(DataAPITime.from_time(other))
        else:
            return NotImplemented

    def _to_tuple(self) -> tuple[int, int, int, int]:
        return (self.hour, self.minute, self.second, self.nanosecond)

    def to_string(self) -> str:
        """
        Convert the DataAPITime to a string according to the Data API specification.

        Returns:
            a string expressing the time value in a way compatible with the Data API
            conventions.

        Example:
            >>> from astrapy.data_types import DataAPITime
            >>>
            >>> DataAPITime(12, 34, 56).to_string()
            '12:34:56'
            >>> DataAPITime(21, 43, 56, 789012345).to_string()
            '21:43:56.789012345'
        """

        hm_part = f"{self.hour:02}:{self.minute:02}"
        s_part: str
        if self.nanosecond:
            nano_div: int
            nano_digits: int
            if self.nanosecond % 1000000 == 0:
                nano_div = 1000000
                nano_digits = 3
            elif self.nanosecond % 1000 == 0:
                nano_div = 1000
                nano_digits = 6
            else:
                nano_div = 1
                nano_digits = 9
            ns_format_string = f"%0{nano_digits}i"
            s_part = (
                f"{self.second:02}.{ns_format_string % (self.nanosecond // nano_div)}"
            )
        else:
            s_part = f"{self.second:02}"
        return f"{hm_part}:{s_part}"

    def to_time(self) -> datetime.time:
        """
        Convert the DataAPITime into a Python standard-library `datetime.time` object.
        This may involve a loss of precision since the latter cannot express
        nanoseconds, only microseconds.

        Returns:
            a `datetime.time` object, corresponding (possibly in a lossy way)
            to the original DataAPITime.

        Example:
            >>> from astrapy.data_types import DataAPITime
            >>>
            >>> DataAPITime(12, 34, 56).to_time()
            datetime.time(12, 34, 56)
            >>> DataAPITime(21, 43, 56, 789012345).to_time()
            datetime.time(21, 43, 56, 789012)
        """

        return datetime.time(
            hour=self.hour,
            minute=self.minute,
            second=self.second,
            microsecond=int(self.nanosecond / 1000.0),
        )

    @staticmethod
    def from_time(dt: datetime.time) -> DataAPITime:
        """
        Create a DataAPITime from a Python standard-library `datetime.time` object.

        Args:
            dt: a `datetime.time` value

        Returns:
            a DataAPITime object, corresponding exactly to the provided input.

        Example:
            >>> from datetime import time
            >>>
            >>> from astrapy.data_types import DataAPITime
            >>>
            >>> DataAPITime.from_time(time(12, 34, 56))
            DataAPITime(12, 34, 56, 0)
            >>> DataAPITime.from_time(time(12, 34, 56, 789012))
            DataAPITime(12, 34, 56, 789012000)
        """
        return DataAPITime(
            hour=dt.hour,
            minute=dt.minute,
            second=dt.second,
            nanosecond=dt.microsecond * 1000,
        )

    @staticmethod
    def from_string(time_string: str) -> DataAPITime:
        """
        Parse a string, expressed according to the Data API format, into a DataAPITime.
        If the format is not recognized, a ValueError is raised.

        Args:
            time_string: a string compliant to the Data API specification for times.

        Returns:
            a DataAPITime corresponding to the provided input.

        Example:
            >>> from astrapy.data_types import DataAPITime
            >>>
            >>> DataAPITime.from_string("12:34:56")
            DataAPITime(12, 34, 56, 0)
            >>> DataAPITime.from_string("21:43:56.789012345")
            DataAPITime(21, 43, 56, 789012345)
            >>> DataAPITime.from_string("34:11:22.123")
            Traceback (most recent call last):
              [...]
            ValueError: Cannot parse '34:11:22.123' into a valid time: illegal hour. ...
        """

        match = TIME_PARSE_PATTERN.match(time_string)
        if match:
            hour = int(match[1])
            minute = int(match[2])
            second: int
            nanosecond: int
            if match[3]:
                second_match = SECONDS_PARSE_PATTERN.match(match[3])
                if second_match:
                    second = int(second_match[1])
                    if second_match[2]:
                        nanosecond = int(float(second_match[2]) * 1000000000)
                    else:
                        nanosecond = 0
                else:
                    raise ValueError(
                        f"Cannot parse '{time_string}' into a valid time "
                        f"(unrecognized format). {TIME_FORMAT_DESC}"
                    )
            else:
                second = 0
                nanosecond = 0
            _fail_reason = _validate_time(
                hour=hour,
                minute=minute,
                second=second,
                nanosecond=nanosecond,
            )
            if _fail_reason:
                raise ValueError(
                    f"Cannot parse '{time_string}' into a valid time: "
                    f"{_fail_reason}. {TIME_FORMAT_DESC}"
                )
            return DataAPITime(
                hour=hour,
                minute=minute,
                second=second,
                nanosecond=nanosecond,
            )
        else:
            raise ValueError(
                f"Cannot parse '{time_string}' into a valid time "
                f"(unrecognized format). {TIME_FORMAT_DESC}"
            )



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from astrapy.data.cursors.cursor import (
    AbstractCursor,
    CursorState,
)
from astrapy.data.cursors.farr_cursor import (
    AsyncCollectionFindAndRerankCursor,
    CollectionFindAndRerankCursor,
)
from astrapy.data.cursors.find_cursor import (
    AsyncCollectionFindCursor,
    AsyncTableFindCursor,
    CollectionFindCursor,
    TableFindCursor,
)
from astrapy.data.cursors.reranked_result import RerankedResult

__all__ = [
    "AsyncCollectionFindAndRerankCursor",
    "AsyncCollectionFindCursor",
    "AsyncTableFindCursor",
    "CollectionFindAndRerankCursor",
    "CollectionFindCursor",
    "AbstractCursor",
    "CursorState",
    "RerankedResult",
    "TableFindCursor",
]



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import time
from dataclasses import dataclass

import httpx

from astrapy.exceptions.collection_exceptions import (
    CollectionDeleteManyException,
    CollectionInsertManyException,
    CollectionUpdateManyException,
    TooManyDocumentsToCountException,
)
from astrapy.exceptions.data_api_exceptions import (
    CursorException,
    DataAPIErrorDescriptor,
    DataAPIException,
    DataAPIHttpException,
    DataAPIResponseException,
    DataAPITimeoutException,
    UnexpectedDataAPIResponseException,
)
from astrapy.exceptions.devops_api_exceptions import (
    DevOpsAPIErrorDescriptor,
    DevOpsAPIException,
    DevOpsAPIHttpException,
    DevOpsAPIResponseException,
    DevOpsAPITimeoutException,
    UnexpectedDevOpsAPIResponseException,
)
from astrapy.exceptions.table_exceptions import (
    TableInsertManyException,
    TooManyRowsToCountException,
)
from astrapy.utils.api_options import FullTimeoutOptions


def _min_labeled_timeout(
    *timeouts: tuple[int | None, str | None],
) -> tuple[int, str | None]:
    _non_null: list[tuple[int, str | None]] = [
        to  # type: ignore[misc]
        for to in timeouts
        if to[0] is not None
    ]
    if _non_null:
        min_to, min_lb = min(_non_null, key=lambda p: p[0])
        # min_to is never None, this is for added robustness
        return (min_to or 0, min_lb)
    else:
        return (0, None)


def _select_singlereq_timeout_gm(
    *,
    timeout_options: FullTimeoutOptions,
    general_method_timeout_ms: int | None,
    request_timeout_ms: int | None = None,
    timeout_ms: int | None = None,
) -> tuple[int, str | None]:
    """
    Apply the logic for determining and labeling the timeout for
    (non-admin) single-request methods.

    If no int args are passed, pick (and correctly label) the least of the
    involved parameters.
    If any of the int args are passed, pick (and correctly labeled) the least
    of them, disregarding the options altogether.
    """
    if all(
        iarg is None
        for iarg in (general_method_timeout_ms, request_timeout_ms, timeout_ms)
    ):
        ao_r = timeout_options.request_timeout_ms
        ao_gm = timeout_options.general_method_timeout_ms
        if ao_r < ao_gm:
            return (ao_r, "request_timeout_ms")
        else:
            return (ao_gm, "general_method_timeout_ms")
    else:
        return _min_labeled_timeout(
            (general_method_timeout_ms, "general_method_timeout_ms"),
            (request_timeout_ms, "request_timeout_ms"),
            (timeout_ms, "timeout_ms"),
        )


def _select_singlereq_timeout_ca(
    *,
    timeout_options: FullTimeoutOptions,
    collection_admin_timeout_ms: int | None,
    request_timeout_ms: int | None = None,
    timeout_ms: int | None = None,
) -> tuple[int, str | None]:
    """
    Apply the logic for determining and labeling the timeout for
    (collection-admin) single-request methods.

    If no int args are passed, pick (and correctly label) the least of the
    involved parameters.
    If any of the int args are passed, pick (and correctly labeled) the least
    of them, disregarding the options altogether.
    """
    if all(
        iarg is None
        for iarg in (collection_admin_timeout_ms, request_timeout_ms, timeout_ms)
    ):
        ao_r = timeout_options.request_timeout_ms
        ao_gm = timeout_options.collection_admin_timeout_ms
        if ao_r < ao_gm:
            return (ao_r, "request_timeout_ms")
        else:
            return (ao_gm, "collection_admin_timeout_ms")
    else:
        return _min_labeled_timeout(
            (collection_admin_timeout_ms, "collection_admin_timeout_ms"),
            (request_timeout_ms, "request_timeout_ms"),
            (timeout_ms, "timeout_ms"),
        )


def _select_singlereq_timeout_ta(
    *,
    timeout_options: FullTimeoutOptions,
    table_admin_timeout_ms: int | None,
    request_timeout_ms: int | None = None,
    timeout_ms: int | None = None,
) -> tuple[int, str | None]:
    """
    Apply the logic for determining and labeling the timeout for
    (table-admin) single-request methods.

    If no int args are passed, pick (and correctly label) the least of the
    involved parameters.
    If any of the int args are passed, pick (and correctly labeled) the least
    of them, disregarding the options altogether.
    """
    if all(
        iarg is None
        for iarg in (table_admin_timeout_ms, request_timeout_ms, timeout_ms)
    ):
        ao_r = timeout_options.request_timeout_ms
        ao_gm = timeout_options.table_admin_timeout_ms
        if ao_r < ao_gm:
            return (ao_r, "request_timeout_ms")
        else:
            return (ao_gm, "table_admin_timeout_ms")
    else:
        return _min_labeled_timeout(
            (table_admin_timeout_ms, "table_admin_timeout_ms"),
            (request_timeout_ms, "request_timeout_ms"),
            (timeout_ms, "timeout_ms"),
        )


def _select_singlereq_timeout_da(
    *,
    timeout_options: FullTimeoutOptions,
    database_admin_timeout_ms: int | None,
    request_timeout_ms: int | None = None,
    timeout_ms: int | None = None,
) -> tuple[int, str | None]:
    """
    Apply the logic for determining and labeling the timeout for
    (database-admin) single-request methods.

    If no int args are passed, pick (and correctly label) the least of the
    involved parameters.
    If any of the int args are passed, pick (and correctly labeled) the least
    of them, disregarding the options altogether.
    """
    if all(
        iarg is None
        for iarg in (database_admin_timeout_ms, request_timeout_ms, timeout_ms)
    ):
        ao_r = timeout_options.request_timeout_ms
        ao_gm = timeout_options.database_admin_timeout_ms
        if ao_r < ao_gm:
            return (ao_r, "request_timeout_ms")
        else:
            return (ao_gm, "database_admin_timeout_ms")
    else:
        return _min_labeled_timeout(
            (database_admin_timeout_ms, "database_admin_timeout_ms"),
            (request_timeout_ms, "request_timeout_ms"),
            (timeout_ms, "timeout_ms"),
        )


def _select_singlereq_timeout_ka(
    *,
    timeout_options: FullTimeoutOptions,
    keyspace_admin_timeout_ms: int | None,
    request_timeout_ms: int | None = None,
    timeout_ms: int | None = None,
) -> tuple[int, str | None]:
    """
    Apply the logic for determining and labeling the timeout for
    (keyspace-admin) single-request methods.

    If no int args are passed, pick (and correctly label) the least of the
    involved parameters.
    If any of the int args are passed, pick (and correctly labeled) the least
    of them, disregarding the options altogether.
    """
    if all(
        iarg is None
        for iarg in (keyspace_admin_timeout_ms, request_timeout_ms, timeout_ms)
    ):
        ao_r = timeout_options.request_timeout_ms
        ao_gm = timeout_options.keyspace_admin_timeout_ms
        if ao_r < ao_gm:
            return (ao_r, "request_timeout_ms")
        else:
            return (ao_gm, "keyspace_admin_timeout_ms")
    else:
        return _min_labeled_timeout(
            (keyspace_admin_timeout_ms, "keyspace_admin_timeout_ms"),
            (request_timeout_ms, "request_timeout_ms"),
            (timeout_ms, "timeout_ms"),
        )


def _first_valid_timeout(
    *items: tuple[int | None, str | None],
) -> tuple[int, str | None]:
    # items are: (timeout ms, label)
    not_nulls = [itm for itm in items if itm[0] is not None]
    if not_nulls:
        return not_nulls[0]  # type: ignore[return-value]
    else:
        # If no non-nulls provided, return zero
        # (a timeout of zero will stand for 'no timeout' later on to the request).
        return 0, None


class InvalidEnvironmentException(Exception):
    """
    An operation was attempted, that is not available on the specified
    environment. For example, trying to get an AstraDBAdmin from a client
    set to a non-Astra environment.
    """

    pass


def to_dataapi_timeout_exception(
    httpx_timeout: httpx.TimeoutException,
    timeout_context: _TimeoutContext,
) -> DataAPITimeoutException:
    text: str
    text_0 = str(httpx_timeout) or "timed out"
    timeout_ms = timeout_context.nominal_ms or timeout_context.request_ms
    timeout_label = timeout_context.label
    if timeout_ms:
        if timeout_label:
            text = f"{text_0} (timeout honoured: {timeout_label} = {timeout_ms} ms)"
        else:
            text = f"{text_0} (timeout honoured: {timeout_ms} ms)"
    else:
        text = text_0
    if isinstance(httpx_timeout, httpx.ConnectTimeout):
        timeout_type = "connect"
    elif isinstance(httpx_timeout, httpx.ReadTimeout):
        timeout_type = "read"
    elif isinstance(httpx_timeout, httpx.WriteTimeout):
        timeout_type = "write"
    elif isinstance(httpx_timeout, httpx.PoolTimeout):
        timeout_type = "pool"
    else:
        timeout_type = "generic"
    if httpx_timeout.request:
        endpoint = str(httpx_timeout.request.url)
        if isinstance(httpx_timeout.request.content, bytes):
            raw_payload = httpx_timeout.request.content.decode()
        else:
            raw_payload = None
    else:
        endpoint = None
        raw_payload = None
    return DataAPITimeoutException(
        text=text,
        timeout_type=timeout_type,
        endpoint=endpoint,
        raw_payload=raw_payload,
    )


def to_devopsapi_timeout_exception(
    httpx_timeout: httpx.TimeoutException,
    timeout_context: _TimeoutContext,
) -> DevOpsAPITimeoutException:
    text: str
    text_0 = str(httpx_timeout) or "timed out"
    timeout_ms = timeout_context.nominal_ms or timeout_context.request_ms
    timeout_label = timeout_context.label
    if timeout_ms:
        if timeout_label:
            text = f"{text_0} (timeout honoured: {timeout_label} = {timeout_ms} ms)"
        else:
            text = f"{text_0} (timeout honoured: {timeout_ms} ms)"
    if isinstance(httpx_timeout, httpx.ConnectTimeout):
        timeout_type = "connect"
    elif isinstance(httpx_timeout, httpx.ReadTimeout):
        timeout_type = "read"
    elif isinstance(httpx_timeout, httpx.WriteTimeout):
        timeout_type = "write"
    elif isinstance(httpx_timeout, httpx.PoolTimeout):
        timeout_type = "pool"
    else:
        timeout_type = "generic"
    if httpx_timeout.request:
        endpoint = str(httpx_timeout.request.url)
        if isinstance(httpx_timeout.request.content, bytes):
            raw_payload = httpx_timeout.request.content.decode()
        else:
            raw_payload = None
    else:
        endpoint = None
        raw_payload = None
    return DevOpsAPITimeoutException(
        text=text,
        timeout_type=timeout_type,
        endpoint=endpoint,
        raw_payload=raw_payload,
    )


@dataclass
class _TimeoutContext:
    """
    This class encodes standardized "enriched information" attached to a timeout
    value to obey. This makes it possible, in case the timeout is raised, to present
    the user with a better error message detailing the name of the setting responsible
    for the timeout and the "nominal" value (which may not always coincide with the
    actual elapsed number of milliseconds because of cumulative timeouts spanning
    several HTTP requests).

    Args:
        nominal_ms: the original timeout in milliseconds that was ultimately set by
            the user.
        request_ms: the actual number of millisecond a given HTTP request was allowed
            to last. This may be smaller than `nominal_ms` because of timeouts imposed
            on a succession of requests.
        label: a string, providing the name of the timeout setting as known by the user.
    """

    nominal_ms: int | None
    request_ms: int | None
    label: str | None

    def __init__(
        self,
        *,
        request_ms: int | None,
        nominal_ms: int | None = None,
        label: str | None = None,
    ) -> None:
        self.nominal_ms = nominal_ms
        self.request_ms = request_ms
        self.label = label

    def __bool__(self) -> bool:
        return self.nominal_ms is not None or self.request_ms is not None


class MultiCallTimeoutManager:
    """
    A helper class to keep track of timing and timeouts
    in a multi-call method context.

    Args:
        overall_timeout_ms: an optional max duration to track (milliseconds)
        dev_ops_api: whether this manager is used in a DevOps context (a fact
            which affects which timeout exception is raised if needed).

    Attributes:
        overall_timeout_ms: an optional max duration to track (milliseconds)
        started_ms: timestamp of the instance construction (milliseconds)
        deadline_ms: optional deadline in milliseconds (computed by the class).
        timeout_label: a string label identifying the `overall_timeout_ms` in a way
            that is understood by the user who can set timeouts.
    """

    overall_timeout_ms: int | None
    started_ms: int = -1
    deadline_ms: int | None
    timeout_label: str | None

    def __init__(
        self,
        overall_timeout_ms: int | None,
        dev_ops_api: bool = False,
        timeout_label: str | None = None,
    ) -> None:
        self.started_ms = int(time.time() * 1000)
        self.timeout_label = timeout_label
        # zero timeouts provided internally are mapped to None for deadline mgmt:
        self.overall_timeout_ms = overall_timeout_ms or None
        self.dev_ops_api = dev_ops_api
        if self.overall_timeout_ms is not None:
            self.deadline_ms = self.started_ms + self.overall_timeout_ms
        else:
            self.deadline_ms = None

    def remaining_timeout(
        self, cap_time_ms: int | None = None, cap_timeout_label: str | None = None
    ) -> _TimeoutContext:
        """
        Ensure the deadline, if any, is not yet in the past.
        If it is, raise an appropriate timeout error.
        If not, return either None (if no timeout) or the remaining milliseconds.
        For use within the multi-call method.

        Args:
            cap_time_ms: an additional timeout constraint to cap the result of
                this method. If the remaining timeout from this manager exceeds
                the provided cap, the cap is returned.
            cap_timeout_label: the label identifying the "cap timeout" if one is
                set. This is for the purpose of tracing the "lineage" of timeout
                settings in a way that is transparent to the user.

        Returns:
            A _TimeoutContext appropriately detailing the residual time an overall
            operation is allowed to last. Alternatively, the method may not return
            and raise a DataAPITimeoutException/DevOpsAPITimeoutException directly.
        """

        # a zero 'cap' must be treated as None:
        _cap_time_ms = cap_time_ms or None
        now_ms = int(time.time() * 1000)
        if self.deadline_ms is not None:
            if now_ms < self.deadline_ms:
                remaining = self.deadline_ms - now_ms
                if _cap_time_ms is None:
                    return _TimeoutContext(
                        nominal_ms=self.overall_timeout_ms,
                        request_ms=remaining,
                        label=self.timeout_label,
                    )
                else:
                    if remaining > _cap_time_ms:
                        return _TimeoutContext(
                            nominal_ms=_cap_time_ms,
                            request_ms=_cap_time_ms,
                            label=cap_timeout_label,
                        )
                    else:
                        return _TimeoutContext(
                            nominal_ms=self.overall_timeout_ms,
                            request_ms=remaining,
                            label=self.timeout_label,
                        )
            else:
                err_msg: str
                if self.timeout_label:
                    err_msg = (
                        f"Operation timed out (timeout honoured: {self.timeout_label} "
                        f"= {self.overall_timeout_ms} ms)."
                    )
                else:
                    err_msg = f"Operation timed out (timeout honoured: {self.overall_timeout_ms} ms)."
                if not self.dev_ops_api:
                    raise DataAPITimeoutException(
                        text=err_msg,
                        timeout_type="generic",
                        endpoint=None,
                        raw_payload=None,
                    )
                else:
                    raise DevOpsAPITimeoutException(
                        text=err_msg,
                        timeout_type="generic",
                        endpoint=None,
                        raw_payload=None,
                    )
        else:
            if _cap_time_ms is None:
                return _TimeoutContext(
                    nominal_ms=self.overall_timeout_ms,
                    request_ms=None,
                    label=self.timeout_label,
                )
            else:
                return _TimeoutContext(
                    nominal_ms=_cap_time_ms,
                    request_ms=_cap_time_ms,
                    label=cap_timeout_label,
                )


__all__ = [
    "DevOpsAPIException",
    "DevOpsAPIHttpException",
    "DevOpsAPITimeoutException",
    "DevOpsAPIErrorDescriptor",
    "UnexpectedDevOpsAPIResponseException",
    "DevOpsAPIResponseException",
    "DataAPIErrorDescriptor",
    "DataAPIDetailedErrorDescriptor",
    "DataAPIException",
    "DataAPIHttpException",
    "DataAPITimeoutException",
    "CursorException",
    "TooManyDocumentsToCountException",
    "UnexpectedDataAPIResponseException",
    "DataAPIResponseException",
    "CollectionInsertManyException",
    "CollectionDeleteManyException",
    "CollectionUpdateManyException",
    "MultiCallTimeoutManager",
    "TooManyRowsToCountException",
    "TableInsertManyException",
]

__pdoc__ = {
    "to_dataapi_timeout_exception": False,
    "MultiCallTimeoutManager": False,
}



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import datetime
from dataclasses import dataclass

from astrapy.utils.duration_c_utils import (
    _build_c_duration_string,
    _parse_c_duration_string,
)
from astrapy.utils.duration_std_utils import (
    _build_std_duration_string,
    _parse_std_duration_string,
)


@dataclass
class DataAPIDuration:
    """
    A "duration" value, suitable for working with the "duration" table column type.

    Durations are an abstract notion: as such, they cannot be mapped into a precise,
    well-defined time span. In other words, their components cannot be referred
    one to another. Indeed, a 'month' is not a fixed amount of days, and likewise
    a 'day' is not a fixed amount of hours (due to daylight saving changes, there are
    days made of 23 or 25 hours).

    For this reason, the DataAPIDuration class cannot really be mapped to a Python
    standard-library `datetime.timedelta`, unless in a restricted set of cases and/or
    with some approximations involved.

    Args:
        months: an integer non-negative amount of months.
        days: an integer non-negative amount of days.
        nanoseconds: an integer non-negative amount of nanoseconds.
            This quantity encodes the whole sub-day component of the
            duration: for instance, a duration of 36 hours is expressed as
            36 * 3600 * 1000000000 = 129600000000000 nanoseconds.
        signum: an overall plus or minus sign, represented as either +1 or -1.
            This allows the encoding of negative durations.
    """

    months: int
    days: int
    nanoseconds: int
    signum: int

    def __init__(
        self,
        signum: int,
        months: int,
        days: int,
        nanoseconds: int,
    ) -> None:
        if months < 0 or days < 0 or nanoseconds < 0:
            raise ValueError(
                "months, days, nanoseconds cannot be negative. Use overall 'signum'."
            )
        if signum not in {+1, -1}:
            raise ValueError("signum must be either +1 or -1.")
        self.months = months
        self.days = days
        self.nanoseconds = nanoseconds
        self.signum = signum

    def __repr__(self) -> str:
        def irepr(val: int) -> str:
            if val != 0 and self.signum < 0:
                return f"-{val}"
            else:
                return f"{val}"

        inner_desc = (
            f"months={irepr(self.months)}, days={irepr(self.days)}, "
            f"nanoseconds={irepr(self.nanoseconds)}"
        )
        return f"{self.__class__.__name__}({inner_desc})"

    def __str__(self) -> str:
        return self.to_string()

    def __reduce__(self) -> tuple[type, tuple[int, int, int, int]]:
        return self.__class__, (self.signum, self.months, self.days, self.nanoseconds)

    def __hash__(self) -> int:
        return hash((self.signum, self.months, self.days, self.nanoseconds))

    @staticmethod
    def from_string(duration_string: str) -> DataAPIDuration:
        """
        Parse a string, expressed according to the Data API ISO-8601 format, into a
        DataAPIDuration. If the format is not recognized, a ValueError is raised.

        Args:
            duration_string: a string compliant to the Data API ISO-8601 specification
            for durations.

        Returns:
            a DataAPIDuration corresponding to the provided input.

        Example:
            >>> from astrapy.data_types import DataAPIDuration
            >>>
            >>> DataAPIDuration.from_string("P3Y6M4DT12H30M5S")
            DataAPIDuration(months=42, days=4, nanoseconds=45005000000000)
            >>> DataAPIDuration.from_string("PT12H")
            DataAPIDuration(months=0, days=0, nanoseconds=43200000000000)
            >>> DataAPIDuration.from_string("PT12H11X")
            Traceback (most recent call last):
              [...]
            ValueError: Invalid fraction-of-day component for a duration string ...
        """

        si, mo, da, ns = _parse_std_duration_string(duration_string)
        return DataAPIDuration(
            signum=si,
            months=mo,
            days=da,
            nanoseconds=ns,
        )

    def to_string(self) -> str:
        """
        Convert the DataAPIDuration to a string according to the Data API
        ISO-8601 standard.

        Returns:
            a string expressing the time value in a way compatible with the Data API
            conventions.

        Example:
            >>> from astrapy.data_types import DataAPIDuration
            >>>
            >>> dd1 = DataAPIDuration(1, months=42, days=4, nanoseconds=45005000000000)
            >>> dd1.to_string()
            'P3Y6M4DT12H30M5S'
            >>> dd2 = DataAPIDuration(1, months=0, days=0, nanoseconds=43200000000000)
            >>> dd2.to_string()
            'PT12H'
        """

        return _build_std_duration_string(
            signum=self.signum,
            months=self.months,
            days=self.days,
            nanoseconds=self.nanoseconds,
        )

    @staticmethod
    def from_c_string(duration_string: str) -> DataAPIDuration:
        """
        Parse a string, expressed according to the Data API "Apache Cassandra(R)
        notation", into a DataAPIDuration.
        If the format is not recognized, a ValueError is raised.

        Args:
            duration_string: a string compliant to the Data API "Apache Cassandra(R)
            notation" for durations.

        Returns:
            a DataAPIDuration corresponding to the provided input.

        Example:
            >>> from astrapy.data_types import DataAPIDuration
            >>>
            >>> DataAPIDuration.from_c_string("12y3mo1d")
            DataAPIDuration(months=147, days=1, nanoseconds=0)
            >>> DataAPIDuration.from_c_string("12y3mo1d12h30m5s")
            DataAPIDuration(months=147, days=1, nanoseconds=45005000000000)
            >>> DataAPIDuration.from_c_string("-4h1978us")
            DataAPIDuration(months=0, days=0, nanoseconds=-14400001978000)
            >>> DataAPIDuration.from_c_string("0h0m")
            DataAPIDuration(months=0, days=0, nanoseconds=0)
            >>> DataAPIDuration.from_c_string("1h1y")
            Traceback (most recent call last):
              [...]
            ValueError: Unit 'y' cannot follow smaller units in literal for ...
        """

        si, mo, da, ns = _parse_c_duration_string(duration_string)
        return DataAPIDuration(
            signum=si,
            months=mo,
            days=da,
            nanoseconds=ns,
        )

    def to_c_string(self) -> str:
        """
        Convert the DataAPIDuration to a string according to the Data API
        "Apache Cassandra(R) notation".

        Returns:
            a string expressing the time value in a way compatible with the Data API
            conventions.

        Example:
            >>> from astrapy.data_types import DataAPIDuration
            >>>
            >>> dd1 = DataAPIDuration(1, months=147, days=1, nanoseconds=0)
            >>> dd1.to_c_string()
            '12y3mo1d'
            >>> dd2 = DataAPIDuration(1, months=147, days=1, nanoseconds=45005000000000)
            >>> dd2.to_c_string()
            '12y3mo1d12h30m5s'
            >>> dd3 = DataAPIDuration(-1, months=0, days=0, nanoseconds=14400001978000)
            >>> dd3.to_c_string()
            '-4h1ms978us'
            >>> dd4 = DataAPIDuration(1, months=0, days=0, nanoseconds=0)
            >>> dd4.to_c_string()
            '0s'
        """

        return _build_c_duration_string(
            signum=self.signum,
            months=self.months,
            days=self.days,
            nanoseconds=self.nanoseconds,
        )

    @staticmethod
    def from_timedelta(td: datetime.timedelta) -> DataAPIDuration:
        """
        Construct a DataAPIDuration from a Python standard-library `datetime.timedelta`.

        Due to the intrinsic difference between the notions of a DataAPIDuration and
        the timedelta, the latter - a definite time span - is only ever resulting
        in a duration with a nonzero "nanoseconds" component.

        Args:
            dt: a `datetime.timedelta` value.

        Returns:
            A DataAPIDuration corresponding value, with null month and day components
            by construction.

        Example:
            >>> from datetime import timedelta
            >>>
            >>> from astrapy.data_types import DataAPIDuration
            >>>
            >>> DataAPIDuration.from_timedelta(
            ...     timedelta(days=1, hours=2, seconds=10.987)
            ... )
            DataAPIDuration(months=0, days=0, nanoseconds=93610987000000)
            >>> DataAPIDuration.from_timedelta(timedelta(hours=-4))
            DataAPIDuration(months=0, days=0, nanoseconds=-14400000000000)
        """

        # this conversion expresses a duration with sub-days component only,
        # since a 'timedelta' is a precise time span (as opposed to durations).
        total_nanoseconds = int(td.total_seconds() * 1000000000)
        if total_nanoseconds >= 0:
            return DataAPIDuration(
                signum=+1,
                months=0,
                days=0,
                nanoseconds=total_nanoseconds,
            )
        else:
            return DataAPIDuration(
                signum=-1,
                months=0,
                days=0,
                nanoseconds=-total_nanoseconds,
            )

    def to_timedelta(self) -> datetime.timedelta:
        """
        Convert a DataAPIDuration into a Python standard library `datetime.timedelta`.

        Due to the intrinsic difference between the notions of a DataAPIDuration and
        the timedelta, the conversion attempt raises an error if the starting
        DataAPIDuration has a nonzero number of months. For the same reason, a somewhat
        lossy conversion occurs for a nonzero number of days since the formal notion
        of a day is lost in favor of that of "a span of exactly 24 hours".

        Returns:
            a `datetime.timedelta` value corresponding to the origin DataAPIDuration
            in the best possible way.

        Example:
            >>> from astrapy.data_types import DataAPIDuration
            >>>
            >>> dd1 = DataAPIDuration(
            ...     signum=-1, months=0, days=0, nanoseconds=93610987000000
            ... )
            >>> dd1.to_timedelta()
            datetime.timedelta(days=-2, seconds=79189, microseconds=13000)
            >>> dd2 = DataAPIDuration(
            ...     signum=1, months=0, days=0, nanoseconds=14400000000000
            ... )
            >>> dd2.to_timedelta()
            datetime.timedelta(seconds=14400)
            >>> dd3 = DataAPIDuration(signum=1, months=19, days=0, nanoseconds=0)
            >>> dd3.to_timedelta()
            Traceback (most recent call last):
                [...]
            ValueError: Cannot convert a DataAPIDuration with nonzero months into ...
        """

        if self.months != 0:
            raise ValueError(
                "Cannot convert a DataAPIDuration with nonzero months into a timedelta."
            )
        return datetime.timedelta(
            days=self.signum * self.days,
            microseconds=self.signum * self.nanoseconds // 1000,
        )



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from dataclasses import dataclass
from typing import Any

import httpx


class DataAPIException(Exception):
    """
    Any exception occurred while issuing requests to the Data API
    and specific to it, such as:
      - a collection is found not to exist when gettings its metadata,
      - the API return a response with an error,
    but not, for instance,
      - a network error while sending an HTTP request to the API.
    """

    pass


@dataclass
class DataAPIErrorDescriptor:
    """
    An object representing a single error, or warning, as returned from the Data API,
    typically with an error code, a text message and other properties.

    This object is used to describe errors/warnings received from the Data API,
    in the form of HTTP-200 ("success") responses containing errors
    (and possibly warnings).
    Depending on the API command semantics, responses may express partial successes
    with some errors (for instance, an insertMany command inserting most of the
    documents/rows, but failing on a couple of incompatible inputs).

    Attributes:
        error_code: a string code as found in the API "error" item.
        message: the text found in the API "error" item.
        title:  the text found in the API "title" item.
        family:  the text found in the API "family" item.
        scope:  the text found in the API "scope" item.
        id:  the text found in the API "id" item.
        attributes: a dict with any further key-value pairs returned by the API.
    """

    title: str | None
    error_code: str | None
    message: str | None
    family: str | None
    scope: str | None
    id: str | None
    attributes: dict[str, Any]

    _known_dict_fields = {
        "title",
        "errorCode",
        "message",
        "family",
        "scope",
        "id",
    }

    def __init__(self, error_dict: dict[str, str] | str) -> None:
        if isinstance(error_dict, str):
            self.message = error_dict
            self.title = None
            self.error_code = None
            self.family = None
            self.scope = None
            self.id = None
            self.attributes = {}
        else:
            self.title = error_dict.get("title")
            self.error_code = error_dict.get("errorCode")
            self.message = error_dict.get("message")
            self.family = error_dict.get("family")
            self.scope = error_dict.get("scope")
            self.id = error_dict.get("id")
            self.attributes = {
                k: v for k, v in error_dict.items() if k not in self._known_dict_fields
            }

    def __repr__(self) -> str:
        pieces = [
            f"{self.title.__repr__()}" if self.title else None,
            f"error_code={self.error_code.__repr__()}" if self.error_code else None,
            f"message={self.message.__repr__()}" if self.message else None,
            f"family={self.family.__repr__()}" if self.family else None,
            f"scope={self.scope.__repr__()}" if self.scope else None,
            f"id={self.id.__repr__()}" if self.id else None,
            f"attributes={self.attributes.__repr__()}" if self.attributes else None,
        ]
        return f"{self.__class__.__name__}({', '.join(pc for pc in pieces if pc)})"

    def __str__(self) -> str:
        return self.summary()

    def summary(self) -> str:
        """
        Determine a string succinct description of this error descriptor.

        The precise format of this summary is determined by which fields are set.
        """
        non_code_part: str | None
        if self.title:
            if self.message:
                non_code_part = f"{self.title}: {self.message}"
            else:
                non_code_part = f"{self.title}"
        else:
            if self.message:
                non_code_part = f"{self.message}"
            else:
                non_code_part = None
        if self.error_code:
            if non_code_part:
                return f"{non_code_part} ({self.error_code})"
            else:
                return f"{self.error_code}"
        else:
            if non_code_part:
                return non_code_part
            else:
                return ""


DataAPIWarningDescriptor = DataAPIErrorDescriptor


@dataclass
class DataAPIResponseException(DataAPIException):
    """
    The Data API returned an HTTP 200 ("success") response, which however
    reports API-specific error(s), possibly alongside partial successes.

    Attributes:
        text: a text message about the exception.
        command: the payload to the API that led to the response.
        raw_response: the full response from the API.
        error_descriptors: a list of DataAPIErrorDescriptor, one for each
            item in the API response's "errors" field.
        warning_descriptors: a list of DataAPIWarningDescriptor, one for each
            item in the API response's "warnings" field (if there are any).
    """

    text: str | None
    command: dict[str, Any] | None
    raw_response: dict[str, Any]
    error_descriptors: list[DataAPIErrorDescriptor]
    warning_descriptors: list[DataAPIWarningDescriptor]

    def __init__(
        self,
        text: str | None,
        *,
        command: dict[str, Any] | None,
        raw_response: dict[str, Any],
        error_descriptors: list[DataAPIErrorDescriptor],
        warning_descriptors: list[DataAPIWarningDescriptor],
    ) -> None:
        super().__init__(text)
        self.text = text
        self.command = command
        self.raw_response = raw_response
        self.error_descriptors = error_descriptors
        self.warning_descriptors = warning_descriptors

    @staticmethod
    def from_response(
        *,
        command: dict[str, Any] | None,
        raw_response: dict[str, Any],
        **kwargs: Any,
    ) -> DataAPIResponseException:
        """Parse a raw response from the API into this exception."""

        error_descriptors = [
            DataAPIErrorDescriptor(error_dict)
            for error_dict in raw_response.get("errors") or []
        ]
        warning_descriptors = [
            DataAPIWarningDescriptor(error_dict)
            for error_dict in raw_response.get("warnings") or []
        ]

        if error_descriptors:
            summaries = [e_d.summary() for e_d in error_descriptors]
            if len(summaries) == 1:
                text = summaries[0]
            else:
                _j_summaries = "; ".join(
                    f"[{summ_i + 1}] {summ_s}"
                    for summ_i, summ_s in enumerate(summaries)
                )
                text = f"[{len(summaries)} errors collected] {_j_summaries}"
        else:
            text = ""

        return DataAPIResponseException(
            text,
            command=command,
            raw_response=raw_response,
            error_descriptors=error_descriptors,
            warning_descriptors=warning_descriptors,
            **kwargs,
        )


@dataclass
class DataAPIHttpException(DataAPIException, httpx.HTTPStatusError):
    """
    A request to the Data API resulted in an HTTP 4xx or 5xx response.

    In most cases this comes with additional information: the purpose
    of this class is to present such information in a structured way,
    akin to what happens for the DataAPIResponseException, while
    still raising (a subclass of) `httpx.HTTPStatusError`.

    Attributes:
        text: a text message about the exception.
        error_descriptors: a list of all DataAPIErrorDescriptor objects
            found in the response.
    """

    text: str | None
    error_descriptors: list[DataAPIErrorDescriptor]

    def __init__(
        self,
        text: str | None,
        *,
        httpx_error: httpx.HTTPStatusError,
        error_descriptors: list[DataAPIErrorDescriptor],
    ) -> None:
        DataAPIException.__init__(self, text)
        httpx.HTTPStatusError.__init__(
            self,
            message=str(httpx_error),
            request=httpx_error.request,
            response=httpx_error.response,
        )
        self.text = text
        self.httpx_error = httpx_error
        self.error_descriptors = error_descriptors

    def __str__(self) -> str:
        return self.text or str(self.httpx_error)

    @classmethod
    def from_httpx_error(
        cls,
        httpx_error: httpx.HTTPStatusError,
        **kwargs: Any,
    ) -> DataAPIHttpException:
        """Parse a httpx status error into this exception."""

        raw_response: dict[str, Any]
        # the attempt to extract a response structure cannot afford failure.
        try:
            raw_response = httpx_error.response.json()
        except Exception:
            raw_response = {}
        error_descriptors = [
            DataAPIErrorDescriptor(error_dict)
            for error_dict in raw_response.get("errors") or []
        ]
        if error_descriptors:
            text = f"{error_descriptors[0].message}. {str(httpx_error)}"
        else:
            text = str(httpx_error)

        return cls(
            text=text,
            httpx_error=httpx_error,
            error_descriptors=error_descriptors,
            **kwargs,
        )


@dataclass
class DataAPITimeoutException(DataAPIException):
    """
    A Data API operation timed out. This can be a request timeout occurring
    during a specific HTTP request, or can happen over the course of a method
    involving several requests in a row, such as a paginated find.

    Attributes:
        text: a textual description of the error
        timeout_type: this denotes the phase of the HTTP request when the event
            occurred ("connect", "read", "write", "pool") or "generic" if there is
            not a specific request associated to the exception.
        endpoint: if the timeout is tied to a specific request, this is the
            URL that the request was targeting.
        raw_payload:  if the timeout is tied to a specific request, this is the
            associated payload (as a string).
    """

    text: str
    timeout_type: str
    endpoint: str | None
    raw_payload: str | None

    def __init__(
        self,
        text: str,
        *,
        timeout_type: str,
        endpoint: str | None,
        raw_payload: str | None,
    ) -> None:
        super().__init__(text)
        self.text = text
        self.timeout_type = timeout_type
        self.endpoint = endpoint
        self.raw_payload = raw_payload


@dataclass
class CursorException(DataAPIException):
    """
    The cursor operation cannot be invoked if a cursor is not in its pristine
    state (i.e. is already being consumed or is exhausted altogether).

    Attributes:
        text: a text message about the exception.
        cursor_state: a string description of the current state
            of the cursor. See the documentation for Cursor.
    """

    text: str
    cursor_state: str

    def __init__(
        self,
        text: str,
        *,
        cursor_state: str,
    ) -> None:
        super().__init__(text)
        self.text = text
        self.cursor_state = cursor_state


@dataclass
class UnexpectedDataAPIResponseException(DataAPIException):
    """
    The Data API response is malformed in that it does not have
    expected field(s), or they are of the wrong type.

    Attributes:
        text: a text message about the exception.
        raw_response: the response returned by the API in the form of a dict.
    """

    text: str
    raw_response: dict[str, Any] | None

    def __init__(
        self,
        text: str,
        raw_response: dict[str, Any] | None,
    ) -> None:
        super().__init__(text)
        self.text = text
        self.raw_response = raw_response
